#! /usr/bin/env python3
#
# Sanity check for time-tracking report.
#
# Sample input:
#
# Week of Mon 14 Oct 14
# 
# Mon:
# ;; day off
# hours: 0
# 
# Tues:
# 2	3	appserver overview; diagnosing hg update issues on systen repository
# 5	6	re-cloning and reviewing system repository source structure
# 7	8	adding new repository branch; checking old system repository differences (e.g., TODO notes in code)
# 9:30	11:30	exporting related-title code from search prototype into appserver; adding prequisites for running appserver locally (for syntax checking)
# 12	2	trying to get @route-based dispatching for relatedJobs method
# hours: 7
#
# Wed:
# 10    7       [ACME] in-house training for new client
#
#--------------------------------------------------------------------------------
# Notes:
# - The am/pm indicator can be omitted unless ambiguous.
# - The description can start with a tag (e.g., ACME) above for tabulating hours.
#
#-------------------------------------------------------------------------------
# TODO:
# - ** Add edit distance check for 'break' mispellings (e.g., breal) to avoid recording excess time!!
# - Allow _ placeholders in week and total summary if month not complete.
# - Clean up debug_format calls.
# - Make hours line optional (e.g., assume blank line separates days).
# - Make weekly hours optional (e.g., assume repeated day separates week).
# - Have option to output template for current month!
#
#

"""Validates the hour tabulations in a time tracking report"""

# Standard packages
import argparse
from collections import defaultdict
import os
import re
## OLD: import sys

# Local packages
from mezcla import debug
## OLD: import tpo_common as tpo
## OLD: from mezcla import sys_version_info_hack       # pylint: disable=unused-import
from mezcla import system
from mezcla.my_regex import my_re
from mezcla import text_utils

# Note: python 3.6+ format strings are used (n.b., assumes sys_version_info_hack above)
# TODO: See if why to have pylist issue warning about version; that way, there's less
# head scratching about syntax error messages (e.g., for python-3.6+ f-string interpolation).
## TEST:
## assert((sys.version_info.major >= 3) and (sys.version_info.minor >= 6))
## assert(system.python_maj_min_version > 3.6, "Python 3.6+ needed for f-strings")
## if (system.python_maj_min_version > 3.6, "Python 3.6+ needed for f-strings"):
##    assert False, "Python 3.6+ needed for f-strings"
##
## TODO:
## try:
##     debug.trace(4, f"Test of f-strings: major/minor version={system.python_maj_min_version()}")
## except SyntaxError:
##     print("Python 3.6+ needed for f-strings")
## NOTE: effing Python: there should be a from-future style import!
##

WEEKDAYS = ["sunday", "monday", "tuesday", "wednesday", "thursday", "friday", "saturday"]
ABBREVIATED_WEEKDAYS = ["sun", "mon", "tue", "wed", "thu", "fri", "sat"]
## TODO: ALT_ABBREVIATED_WEEKDAYS = ["sun", "mon", "tues", "wed", "thurs", "fri", "sat"]
ALL_WEEKDAYS = WEEKDAYS + ABBREVIATED_WEEKDAYS
REPORT_TAGS = "daily weekly total".split()

#...............................................................................

def show_tagged_hours(tagged_hours_hash, prefix=None):
    """Print report of hours assigned to each task in TAGGED_HOURS_HASH"""
    debug.trace(5, "show_tagged_hours({tagged_hours_hash}, [{prefix}])")
    space = None
    if prefix is None:
        prefix = ""
        space = ""
    else:
        space = " "
    report = ""
    for tag in sorted(tagged_hours_hash):
        if tagged_hours_hash[tag]:
            report += ("\t".join([tag, str(tagged_hours_hash[tag])]) + "\n")
    if report:
        print("{pre}{sp}Hours by tag".format(pre=prefix, sp=space).capitalize())
        print(report.strip("\n"))
    return


def main():
    """Entry point for script"""
    ## OLD: debug.trace(4, f"main(): argv={sys.argv}")
    debug.trace(4, f"main(): argv={system.get_args()}")

    # Check command-line arguments
    # TODO: Switch over to argument processing via Main class in main.py.
    # TODO: Add example illustrating DEBUG_LEVEL and to show check-time-tracking alias;
    # also, add option to output blank template for the month.
    parser = argparse.ArgumentParser(description="Validates time tracking reports")
    parser.add_argument("--strict", default=False, action='store_true', help="strict parsing of the time tracking report")
    parser.add_argument("--ignore-comments", default=False, action='store_true', help="ignore comments in the time tracking report")
    parser.add_argument("--verbose", default=True, action='store_true', help="verbose output mode (e.g., summary of weekly hours)")
    parser.add_argument("--quiet", dest='verbose', action='store_false', help="non-verbose mode")
    ## OLD: parser.add_argument("--weekly", dest='weekly', default=False, action='store_true', help="show weekly summary of hours")
    parser.add_argument("--weekly", default=False, action='store_true', help="show weekly summary of hours")
    parser.add_argument("--report", default=False, action='store_true', help="Output time sheet report including computed hours")
    ## OLD: parser.add_argument("--heuristics", default=False, dest='heuristics', action='store_true', help="use heuristics such as ignoring time slots labelled as 'break' (n.b., can be confusing without --quiet or debugging enabled)")
    ## OLD: parser.add_argument("--heuristics", default=False, action='store_true', help="use heuristics such as ignoring time slots labelled as 'break' (n.b., can be confusing without --quiet or debugging enabled)")
    parser.add_argument("--heuristics", default=True, action='store_true', help="use heuristics such as ignoring time slots labelled as 'break' (n.b., can be confusing without --quiet or debugging enabled)")
    parser.add_argument("--skip-heuristics", dest='heuristics', action='store_false', help="don't use heuristics for time slot interpretation")
    parser.add_argument("--skip-hours-check", default=False, action='store_true', help="Omits warning discrepancies in hour specified vs. tabulated")
    parser.add_argument("--extra-normalization", default=False, action='store_true', help="Extra normalization (e.g., stripping am/pm")
    # TODO: --tags => --tabulate-tags
    parser.add_argument("--tags", default=False, action='store_true', help="Tabulate hours according to tags (e.g., [ACME] review)")
    parser.add_argument("--tag-filter", default="", help="Filter out specified tags")
    parser.add_argument("--tag-report", default="daily,weekly,total", help="Show tag report types: {{{rt}}}".format(rt=", ".join(REPORT_TAGS)))
    parser.add_argument("--omit-time", default=False, action='store_true', help="Omit timespec from report")
    parser.add_argument("--omit-tags", default=False, action='store_true', help="Omit tag from report (n.b., requires --omit-time")
    ## TODO: parser.add_argument("--xyz", default=False, action='store_true', help="TODO: does xyz, ...")
    ## -or- parser.add_argument("--not-xyz", dest='xyz', action='store_false', help="TODO: does not xyz, ...")
    parser.add_argument("filename", help="input filename")
    args = vars(parser.parse_args())
    debug.trace(5, f"args = {args}")
    filename = args['filename']
    debug.assertion(os.path.exists(filename))
    strict = args['strict']
    ignore_comments = args['ignore_comments'] or (not strict)
    use_heuristics = args['heuristics']
    verbose = args['verbose']
    ## TODO: verbose = args['verbose'] or use_heuristics
    show_weekly = args['weekly']
    skip_hours_check = args['skip_hours_check']
    extra_normalization = args['extra_normalization']
    tabulate_tags = args['tags']
    filter_tags = args['tag_filter']
    report_tags = args['tag_report'].lower()
    debug.assertion(not system.difference(re.split("[, ]+", report_tags),
                                          REPORT_TAGS))
    output_report = args['report']
    filter_regex = None
    omit_timespec = args["omit_time"]
    omit_tags = args["omit_tags"]
    debug.assertion(not (omit_tags and not omit_timespec))

    # Format optional regex for tag filtering
    if filter_tags:
        tag_list = text_utils.extract_string_list(filter_tags)
        if not tag_list:
            system.print_stderr(f"Error extracting tags from '{filter_tags}'")
        else:
            # ex: "\[TPO|OTHER\]
            tag_spec="|".join(tag_list)
            filter_regex = r"\[{spec}\]".format(spec=tag_spec)
            debug.trace(2, f"Filter regex: {filter_regex}")

    # Print header for weekly summary and initialize associated record keeping
    if show_weekly or verbose:
        print("\t".join(ABBREVIATED_WEEKDAYS))
    weekday_hours = defaultdict(str)
    day_of_week = ""
    
    # Scan time listing keeping track of hours running totals
    daily_tagged_hours = defaultdict(int)
    weekly_tagged_hours = defaultdict(int)
    total_tagged_hours = defaultdict(int)
    hours = 0
    last_hours = 0
    weekly_hours = 0
    last_weekly_hours = 0
    total_hours = 0
    line_num = 0
    num_weeks = 0
    for original_line in system.open_file(filename):
        original_line = original_line.strip("\n")
        filtered_original_line = original_line
        line = original_line.lower()
        line_num += 1
        debug.trace(6, f"L{line_num}:\n\t{line}")

        # HACK: Standardize the weekday
        line_save = line
        line = re.sub(r"^(sun|mon|tues|wednes|thurs|fri|satur)day:", r"\1:", line)
        line = re.sub(r"^(tues):", r"tue:", line)
        line = re.sub(r"^(wednes):", r"wed:", line)
        line = re.sub(r"^(thurs):", r"thu:", line)
        line = re.sub(r"^(satur):", r"sat:", line)
        if ((line != line_save) or verbose):
            debug.trace_fmt(6, "After day-of-week normalization for L{n}:\n\t{l}", n=line_num, l=line)

        # Normalize line to facilitate extraction:
        # - treat whitespace delimiters as tabs after start and end time
        # - remove surrounding whitespace, make lowercase
        # - convert tabs to spaces
        # TODO: convert "T1-T2<tab>..." to "T1<tab>T2<tab>..."
        line = line.strip()
        line_save = line
        line = re.sub("\t", " ", line)
        if (re.search(r"^(\d+\S*)\s+(\d+\S*)\s+", line)):
            line = re.sub(r"^(\S+)\s+(\S+)\s+", r"\1\t\2\t", line)
            debug.assertion(not line.endswith("\t"))
            debug.assertion(not re.search(r"\t.*\t.*\t", line))
        if ((line != line_save) or verbose):
            debug.trace_fmt(6, "After normalized time spec. for L{n}:\n\t{l}", n=line_num, l=line)

        # Normalize time specification:
        # - remove am/pm suffixes in time specification
        #   -- done when prior to tab (\t), dash (-) or ": " (colon-space sequence)
        ## OLD: line = re.sub(r"(am|pm)\t", r"\t", line)
        if extra_normalization:
            line_save = line
            line = re.sub(r"([0-9])(am|pm)((-)|(:?\t)|(: ))", r"\1\3", line)
            if ((line != line_save) or verbose):
                debug.trace_fmt(6, "After extra normalization for L{n}:\n\t{l}", n=line_num, l=line)

        # Ignore comment lines (starting either w/ '#' or ';' ignoring leading space)
        if (ignore_comments and re.search(r"^\s*[#;]", line)):
            debug.trace(6, f"Ignoring comment in L{line_num}: {line}")
            continue

        # Replace ": " w/ "<TAB>" prior to task description
        # ex: "4-5: cleaning code" => "4-5<TAB>cleaning code"
        # Note: requires am/pm stripping
        # TODO: fix processing over bad-time-tracking-may21.list
        line_save = line
        line = re.sub(r"^([-:\t 0-9]+[0-9]): ", r"\1\t", line)
        if ((line != line_save) or verbose):
            debug.trace_fmt(6, "After misc. normalized for L{n}:\n\t{l}", n=line_num, l=line)
        
        # Strip trailing comments
        # ex: "Fri:	    ;; 28" => "Fri:"
        # Note: double-comment now required, unless two or more preceding spaces
        if ignore_comments:
            line_save = line
            ## OLD: line = re.sub(r"\s*[#;]).*", "", line)
            ## OLD: line = re.sub(r"\s\s\s([#;]).*", "", line)
            line = re.sub(r"\s\s([#;]).*", "", line)
            line = re.sub(r"([#;])\1.*", "", line)
            if ((line != line_save) or verbose):
                debug.trace(6, f"After trailing comment removal L{line_num}:\n\t{line}")
            if re.search(r"\s[#;]", line):
                debug.trace(4, f"Warning: potential comment at line {line_num}: {line}")
                debug.trace(4, "Precede by 2+ spaces or double the comment char")

        # Trace end version of line
        if verbose:
            debug.trace(6, f"Final normalized L{line_num}:\n\t{line}")

        # Flag ???'s as items to flesh out
        if (verbose and line.find("???") != -1):
            print("TODO: flesh out line %d: %s" % (line_num, line))

        # Apply heuristics for entries to ignore, etc. (e.g., ignore time-slots listed as break's).
        ## OLD: if use_heuristics and re.search(r"\tbreak\s*$", line):
        if use_heuristics and re.search(r"\t\s*break\s*$", line):
            message = "Ignoring break at line {n}: {t}".format(n=line_num, t=line)
            if verbose:
                print(message)
            debug.trace(5, f"Warning: {message}")
            continue
            
        # TEMP HACK: first Check for daily hours (redundant with old code below)
        # TODO: See why not handled properly in time-tracking-mmmYY.template
        # ex: "hours: 8.5"
        if (my_re.search(r"^hours:\s*(\S*)", line)):
            ## OLD: debug.trace(5, "redundant hours check")
            debug.trace_fmt(5, "redundant hours check: day_of_week={dw}", dw=day_of_week)
            specified_hours = system.safe_float(my_re.group(1), 0.0) if my_re.group(1) else 0.0
            ## OLD: if (specified_hours != hours):
            if ((specified_hours != hours) and (not skip_hours_check)):
                system.print_stderr("Error: Discrepancy in hours at line {n}: change {spec} => {calc}?".format(n=line_num, spec=specified_hours, calc=hours))
            else:
                # HACK: pretend user specified tabulated hours (TODO: use 'hours' in weekly_hours update below for clarity)
                ## OLD: specified_hours = hours
                pass
            ## OLD: weekly_hours += specified_hours
            weekly_hours += hours
            debug.assertion((weekday_hours[day_of_week] == ""), f"Missing weekly-hours lines (near line {line_num})?")
            ## OLD: weekday_hours[day_of_week] = specified_hours
            ## OLD2: weekday_hours[day_of_week] = hours
            weekday_hours[day_of_week] += ("" if not weekday_hours[day_of_week] else "-or-")
            weekday_hours[day_of_week] += str(hours)
            last_hours = hours
            hours = 0
            debug.trace_fmt(7, "spec_hours={sh} daily_hours={h} weekly_hours={wh} total_hours={th}, weekday_hours={wdh}", 
                            sh=specified_hours, h=hours, wh=weekly_hours, th=total_hours, wdh=weekday_hours)

            # Show tagged breakdown
            if tabulate_tags:
                if "daily" in report_tags:
                    show_tagged_hours(daily_tagged_hours, "Daily")
                for tag in daily_tagged_hours:
                    weekly_tagged_hours[tag] += daily_tagged_hours[tag]
                    daily_tagged_hours[tag] = 0

        # Check for day of week
        # NOTE: 'day' now stripped above.
        # TODO: make sure 'day' not included in first regex group
        ## if (my_re.search(r"^(\S+)(day)?:\s*$", line)):
        ## OLD: if (my_re.search(r"^(\S+)(day)?:\s*$", line)):
        elif (my_re.search(r"^(\S+)(day)?:\s*$", line)):
            day_of_week = my_re.group(1)
            debug.assertion(len(day_of_week) == 3, f"Bad day of week at line {line_num}")
            debug.trace(6, f"day of week check: day={day_of_week}")
            ## TODO: handle Saturday and Wednesday
            ## TODO: day_of_week = re.sub(r"(ur|nes)?day$", "", my_re.group(1))
            if (day_of_week in ALL_WEEKDAYS):
                if (hours > 0):
                    system.print_stderr("Error: missing 'hours:' line at line {n}".format(n=line_num))
            else:
                system.print_stderr("Error: Invalid day '{d}' at line {n}", d=day_of_week, n=line_num)

        # Check for hours specification
        # ex: "2:30pm	4:30	debugging check_time_tracking.py"
        # ex: "7-9	trying to get SpaCy NER via Anaconda under edgenode and win10
        # Notes:
        # - The am or pm suffix is optional.
        # - If start time is greater than end time, the latter is assumed to be afternoon.
        # - The start and end time can optionally be separate by a dash (e.g, 8-10pm), without intervening spaces (e.g., not 8 - 10pm)
        #
        # TODO: (my_re.search(r"^(\d+)\:?(\d*)(am|pm)?\s+(\d+)\:?(\d*)\s+\S.*", line))
        ## OLD:
        ## elif (my_re.search(r"^(\d+)\:?(\d*)\s+(\d+)\:?(\d*)\s+\S.*", line) or
        ##      my_re.search(r"^(\d+)\:?(\d*)\-(\d+)\:?(\d*)\s+\S.*", line)):
        elif (my_re.search(r"^(\d+)\:?(\d*)\s+(\d+)\:?(\d*)\s+(\S.*)", line) or
              my_re.search(r"^(\d+)\:?(\d*)\-(\d+)\:?(\d*)\s+(\S.*)", line)):
            debug.trace(6, "time check")
            # TODO: (start_hours, start_mins, start_ampm, end_hours, end_mins, start_ampm) = my_re.groups()
            ## OLD: (start_hours, start_mins, end_hours, end_mins) = my_re.groups()
            (start_hours, start_mins, end_hours, end_mins, desc) = my_re.groups()
            debug.trace(6, f"sh={start_hours} sm={start_mins} eh={end_hours} em={end_mins} desc={desc}")
            start_time = float(start_hours) + float(start_mins or 0)/60.0
            end_time = float(end_hours) + float(end_mins or 0)/60.0
            if (end_time < start_time):
                end_time += 12
            new_hours = (end_time - start_time)
            if (not (0 < new_hours <= 24)):
                system.print_stderr("Error: Invalid hour specification at line {n}: calculated as more than a day ({new})!".format(n=line_num, new=new_hours))
            ## OLD:
            ## hours += new_hours
            ## debug.trace(6, f"{new_hours} new hours; hours={hours}")

            # Checked for tagged tasks and record hours separately
            # EX: "[XYZ] proposal revision"
            tag = "n/a"
            if my_re.search(r"\[([a-z0-9_/-]+)\]", desc):
                tag = my_re.group(1)
                debug.trace_fmt(5, "found tag {t} at line {n}", t=tag, n=line_num)
            elif my_re.search(r"\[(\S+)\]", desc):
                debug.trace(4, f"Warning: unrecognizable tag '{my_re.group(1)}' at line {line_num}")
            daily_tagged_hours[tag] += new_hours

            # Ignore specified tags
            if filter_regex and re.search(filter_regex, line, re.IGNORECASE):
                filtered_original_line = None
                debug.trace(2, "Filtered out line {n}: {l}".
                            format(n=line_num, l=original_line))
                debug.trace(4, f"{new_hours} ignored hours; hours={hours}")
            else:
                hours += new_hours
                debug.trace(5, f"{new_hours} new hours; hours={hours}")
                if verbose:
                    filtered_original_line += f" [{new_hours}]"
                
        # Check for daily hours
        # HACK: keep in synch with (temp) hack above
        elif (my_re.search(r"^hours:\s*(\S*)", line)):
            ## OLD: debug.trace(6, "hours check")
            debug.trace_fmt(5, "hours check: day_of_week={dw}", dw=day_of_week)
            specified_hours = system.safe_float(my_re.group(1), 0.0) if my_re.group(1) else 0.0
            ## OLD: if (specified_hours != hours):
            if ((specified_hours != hours) and (not skip_hours_check)):
                system.print_stderr("Error: Discrepancy in hours at line {n}: change {spec} => {calc}?".format(n=line_num, spec=specified_hours, calc=hours))
            else:
                # HACK: pretend user specified tabulated hours (TODO: use 'hours' below for clarify)
                ## specified_hours = hours
                pass
            ## OLD: weekly_hours += specified_hours
            weekly_hours += hours
            debug.assertion(weekday_hours[day_of_week] == "", f"Day of week already used at line {line_num}")
            ## OLD: weekday_hours[day_of_week] = specified_hours
            ## OLD2: weekday_hours[day_of_week] = hours
            weekday_hours[day_of_week] += ("" if not weekday_hours[day_of_week] else "-or-")
            weekday_hours[day_of_week] += str(hours)
            last_hours = hours
            hours = 0
            debug.trace_fmt(7, "spec_hours={sh} daily_hours={h} weekly_hours={wh} total_hours={th}, weekday_hours={wdh}", 
                            sh=specified_hours, h=hours, wh=weekly_hours, th=total_hours, wdh=weekday_hours)

            # Show tagged breakdown
            if tabulate_tags:
                if "daily" in report_tags:
                    show_tagged_hours(daily_tagged_hours, "Daily")
                for tag in daily_tagged_hours:
                    weekly_tagged_hours[tag] += daily_tagged_hours[tag]
                    daily_tagged_hours[tag] = 0
            
        # Validate and reset weekly hours
        elif (my_re.search(r"^weekly hours:\s*(\S*)", line)):
            debug.trace(6, "weekly hours check")
            specified_hours = system.safe_float(my_re.group(1), 0.0) if my_re.group(1) else 0.0
            ## OLD: if (specified_hours != weekly_hours):
            if ((specified_hours != weekly_hours) and (not skip_hours_check)):
                system.print_stderr("Error: Discrepancy in weekly hours at line {n}: change {spec} => {calc}?".format(n=line_num, spec=specified_hours, calc=weekly_hours))
            num_weeks += 1

            # Show hours and reset
            if verbose:
                print("Week %d hours: %s" % (num_weeks, weekly_hours))
            total_hours += weekly_hours
            last_weekly_hours = weekly_hours
            weekly_hours = 0

            # Show hours by task and reset
            if tabulate_tags:
                if "weekly" in report_tags:
                    show_tagged_hours(weekly_tagged_hours, "Weekly")
                for tag in weekly_tagged_hours:
                    total_tagged_hours[tag] += weekly_tagged_hours[tag]
                    weekly_tagged_hours[tag] = 0

            # Show hours by day of week and then reset associated record keeping
            # TODO: use map in place of simple list comprehensions
            if show_weekly or verbose:
                debug.assertion(not system.difference(list(weekday_hours.keys()), ABBREVIATED_WEEKDAYS))
                hour_per_day = [weekday_hours.get(d, "0") for d in ABBREVIATED_WEEKDAYS]
                print("\t".join([system.to_string(h) for h in hour_per_day]))
            weekday_hours = defaultdict(str)
            day_of_week = ""

        # Validate total hours
        elif (my_re.search(r"^total hours:\s*(\S*)", line)):
            debug.trace(6, "total hours check")
            if (weekly_hours != 0):
                system.print_stderr("Error: Missing weekly hours prior to total at line {n}".format(n=line_num))
            specified_hours = system.safe_float(my_re.group(1), 0.0) if my_re.group(1) else 0.0
            ## OLD: if (specified_hours != total_hours):
            if ((specified_hours != total_hours) and (not skip_hours_check)):
                system.print_stderr("Error: Discrepancy in total hours at line {n}: change {spec} => {calc}?".format(n=line_num, spec=specified_hours, calc=total_hours))

        # Ignore miscellaneous line: starts with 5 or more dashes (e.g., "------------------...-----")
        elif (my_re.search(r"^\-\-\-\-\-+", line) or my_re.search(r"^week of", line)):
            debug.trace(6, "miscellaenous line check")

        # Check for blank line without hours specification
        elif (line == ""):
            debug.trace(6, "blank line check")
            if (hours > 0):
                system.print_stderr("Warning: Unexpected blank line at line %d" % line_num)

        # Report lines not recognized
        else:
            system.print_stderr("Warning: Unexpected format at line %d: %s" % (line_num, line))

        # Output report with task details
        if output_report and (filtered_original_line is not None):
            report_line = filtered_original_line.strip()
            if omit_timespec:
                time_regex = r"\d*(:\d*)?(am|pm)?"
                timespec_regex = (time_regex + "-" + time_regex + r"(:|\s)\s*")
                report_line = re.sub(timespec_regex, "- ", report_line)
            if omit_tags:
                report_line = re.sub(r"^- \[\S+\] *", "- ", report_line)
            if re.search(r"^Hours:$", report_line, re.IGNORECASE):
                report_line += f" {last_hours}"
            if re.search(r"^Weekly Hours:$", report_line, re.IGNORECASE):
                report_line += f" {last_weekly_hours}"
            if re.search(r"^Total Hours:$", report_line, re.IGNORECASE):
                if (not verbose):
                    report_line += f" {total_hours}"
            debug.trace(7, f"report_line: {report_line}")
            print(report_line)

    # Do some sanity checks
    if (hours > 0):
        system.print_stderr("Warning: Missing weekly total for last week")
        debug.assertion(not system.difference(list(weekday_hours.keys()), ABBREVIATED_WEEKDAYS))

    # Print summary
    if verbose:
        print("Total hours:  %s" % total_hours)
        weekly_average = (total_hours / num_weeks) if (num_weeks > 0) else total_hours
        print("Average per week: %s" % system.round_num(weekly_average))
        proto_average = (total_hours / (52 / 12.0))
        print("Proto average (i.e., 52/12): %s" % system.round_num(proto_average))

    # Show hours by task
    if tabulate_tags:
        if "total" in report_tags:
            show_tagged_hours(total_tagged_hours, "Total")
    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python3
#
# spell.py: simple spell checker (e.g., via Enchant module)
#
# Note:
# - Requires enchant library package to be installed:
#     apt-get install enchant-2
#   See https://pyenchant.github.io/pyenchant/install.html.
# - New languages can be added as follows:
#     ID=es
#     sudo apt-get install myspell-$ID
#   See https://stackoverflow.com/questions/32712557/pyenchant-italian-and-spanish-languages.
#
# TODO:
# - integrate code for suggestions based on spelling distance
#

"""Basic spell checking"""

# Standard modules
# TODO: sys => system
import sys
import fileinput

# Installed modules
import enchant                  # spell checking

# Local modules
from mezcla import debug
## TODO: from mezcla.main import Main
from mezcla import system
## OLD: from mezcla.my_regex import my_re
from mezcla.text_processing import split_word_tokens


# Environment options
SPELL_LANG = system.getenv_text("SPELL_LANG", "en_US",
                                "Language for spelling")

# Constants
TL = debug.TL

## TEMP:
# pylint: disable=consider-using-f-string

def main():
    """Entry point for script"""
    # Process command line
    # TODO: upgrade to using Main script (see template.py)
    i = 1
    show_usage = (i == len(sys.argv))
    while (i < len(sys.argv)) and (sys.argv[i][0] == "-"):
        if (sys.argv[i] == "--help"):
            show_usage = True
        elif (sys.argv[i] == "-"):
            pass
        else:
            print("Error: unexpected argument '%s'" % sys.argv[i])
            show_usage = True
        i += 1
    if (show_usage):
        print("Usage: %s [options] input-file" % sys.argv[0])
        print("")
        print("Options: [--help]")
        print("")
        print("Examples:")
        print("")
        print("  %s query-keywords.list" % sys.argv[0])
        print("")
        print("  echo 'how now browne cow?' | {prog} -".format(prog=sys.argv[0]))
        print("")
        sys.exit()
    # Discard any used arguments (for sake of fileinput)
    if (i > 1):
        debug.trace(5, f"discarding used args: {sys.argv[1: i]}")
        sys.argv = [sys.argv[0]] + sys.argv[i:]
    
    # Initialize spell checking
    speller = enchant.Dict(SPELL_LANG)
    
    # Check input
    for line in fileinput.input():
        line = line.strip()
        debug.trace_fmt(5, "L{line_num}: {line_text}", line_num=fileinput.filelineno(), line_text=line)
    
        # Extract word tokens and print those not recognized
        ## BAD:
        ## word_tokens = [t.strip() for t in my_re.split(r"\W+", line.lower(), my_re.LOCALE|my_re.UNICODE)
        ##                if (len(t.strip()) > 0)]
        word_tokens = split_word_tokens(line.lower())
        debug.trace(4, f"tokens: {word_tokens}")
        for w in word_tokens:
            if not speller.check(w):
                print(w)

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_DETAILED)
    debug.trace_fmt(TL.USUAL, "Environment options: {eo}",
                    eo=system.formatted_environment_option_descriptions())
    main()



#! /usr/bin/env python3
#
# Miscellaneous tests not tied to particular module.
#
# Note:
# - This is uses to check for enforce some development
#   -- A test exists for each module (e.g., tests/test_fubar.py for ./fubar.py).
#   -- Python files have execute permissions (e.g., chmod ugo+x).
#
# TODO2: Run once a week or so (e.g., to help catch poor test stubs like
# test_train_language_model.py)!
#
# TODO3:
# - Check for common pylint issues (n.b., not nitpicking ones like spacing).
#

"""Miscellaneous/non-module tests"""

# Standard packages
## TODO: from collections import defaultdict

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Constants
LINE_IMPORT_PYDANTIC = "from pydantic import validate_call\n"
## TODO1: don't hardcode /tmp (see below)
OUTPUT_PATH_PYDANTIC = "/tmp/temp_"
DECORATOR_VALIDATION_CALL = "@validate_call\ndef"

class TestMisc(TestWrapper):
    """Class for test case definitions"""
    script_module= None

    def get_python_module_files(self, include_tests=False):
        """Return list of files for python modules, optionally with INCLUDE_TESTS"""
        all_python_modules = gh.run("find . -name '*.py'").splitlines()
        ok_python_modules = []
        for module in all_python_modules:
            dir_path, _filename = system.split_path(module)
            _parent_path, dir_proper = system.split_path(dir_path)
            include = True
            if (dir_proper == "tests"):
                include = include_tests
            if include:
                ok_python_modules.append(module)
        debug.trace_expr(5, ok_python_modules)
        return ok_python_modules
    
    def transform_for_validation(self, file_path):
        """Creates a temporary copy of the script for validation of argument calls (using pydantic)"""
        content = system.read_file(file_path)
        content = my_re.sub(r"^def", r"@validate_call\n\g<0>", content, flags=my_re.MULTILINE)
        ## Uncomment the line below (and comment the line above) if the decorators are previously used
        ## May not be compatible with scripts in mezcla/tests
        # content = my_re.sub(r"^(?:(?!\s*[@'#\']).*?)(\s*)(def )", r'\1@validate_call\n\1\2', content, flags=my_re.MULTILINE)
        ## TODO1:  something like the following (so that shebang line kept as first):
        ## my_re.sub(r"^((from \S+ )?import)", fr"\1{LINE_IMPORT_PYDANTIC}", content, flags=my_re.MULTILINE, count=1)
        content = LINE_IMPORT_PYDANTIC + content
        ## TODO2: use self.get_temp_file(): Lorenzo added new functionality
        output_path = OUTPUT_PATH_PYDANTIC + gh.basename(file_path)
        system.write_file(filename=output_path, text=content)
        return output_path
    
    @pytest.mark.xfail
    def test_01_check_for_tests(self):
        """Make sure test exists for each Python module"""
        debug.trace(4, "test_01_check_for_tests()")
        for module in self.get_python_module_files():
            dir_name, filename = system.split_path(module)
            include = True
            # Check for cases to skip (e.g., __init__.py)
            if not filename.startswith("__"):
                include = False
            # Make sure test file exists
            if include:
                test_path = gh.form_path(dir_name, f"test_{filename}")
                self.do_assert(system.file_exists(test_path),
                               f"module {module} is missing tests")

    @pytest.mark.xfail
    def test_02_check_lib_usages(self):
        """Make sure modules don't use certain standard library calls directly
        Note: The mezcla equivalent should be used for sake of debugging traces"""
        debug.trace(4, "test_02_check_lib_usages()")
        for module in self.get_python_module_files():
            module_code = system.read_file(module)
            bad_usage = False
            # Direct use of sys.stdin.read()
            # TODO3: use bash AST (e.g., to exclude comments)
            if "import sys" in module_code and "sys.stdin.read" in module_code:
                bad_usage = (module != "main.py")
                debug.trace(4, f"Direct use of sys.stdin.read in {module}")
            self.do_assert(not bad_usage, f"module {module} has bad sys.stdin.read usage")

    @pytest.mark.xfail
    def test_03_check_permissions(self):
        """Make sure modules don't use certain standard library calls directly
        Note: The mezcla equivalent should be used for sake of debugging traces"""
        debug.trace(4, "test_03_check_permissions()")
        for module in self.get_python_module_files():
            has_execute_perm = gh.run(f'ls -l "{module}" | grep ^...x..x..x')
            self.do_assert(has_execute_perm, f"module {module} is missing execute permission")

    @pytest.mark.xfail
    def test_04_check_transform_for_validation(self):
        """Make sure the transformation is successful when adding decorators for validation of function calls"""
        debug.trace(4, "test_04_check_transform_for_validation()")

        ## TODO2: use gh.resolve_path
        input_file = "../html_utils.py"
        result = self.transform_for_validation(input_file)
        ## TODO1: don't hardcode refs to /tmp 
        assert "temp" in result and "/tmp/" in result
        content = system.read_file(result)
        ## For debugging: print(content)
        assert content.startswith(LINE_IMPORT_PYDANTIC)
        assert content.count(DECORATOR_VALIDATION_CALL) >= 1

    @pytest.mark.xfail
    def test_05_usage_statements(self):
        """Make sure usage statments refer to valid arguments"""
        debug.trace(4, "test_05_usage_statements()")
        # note: addresses change like --include-header => --header in randomize_lines.py
        #   for module in ...: for usage in module usage: assert (not "test_04_usage_statements" in run_script(usage))
        self.do_assert(False, "TODO: implement")
    
#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Test(s) for ../convert_emoticons.py
#
# Sample input and output (n.b., U+1f638 is grinning-cat character):
#
# - input:
#
#   # Example Input:
#   Nothing to test 😸
#
# - output:
#
#   # Example output:
#   Nothing to test [grinning cat face with smiling eyes]
#

"""Tests for convert_emoticons module"""

# Standard packages
## TODO: from collections import defaultdict

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla.unittest_wrapper import trap_exception
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                  global module object
#    TestTemplate.script_module:  path to file
import mezcla.convert_emoticons as THE_MODULE

# Constants
D = system.path_separator()

class TestIt(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    script_file = TestWrapper.get_module_file_path(__file__)

    @pytest.mark.xfail                   # TODO: remove xfail
    ## TEST: @trap_exception
    def test_over_script(self):
        """Makes sure works as expected over script itself"""
        debug.trace(4, f"TestIt.test_over_script(); self={self}")
        output = self.run_script(options="", data_file=__file__)
        # the example usage should have input emoticon changed to match output
        # note: the example comments use ... for brevity and to simply regex matching
        test_script_contents = system.read_file(__file__)
        debug.trace_expr(6, test_script_contents, max_len=8192)
        char_desc = "grinning cat face with smiling eyes"
        # ex (see above): "# Input:\n#   Nothing to test 😸\n# Output:\n#   Nothing to test [Grinning Cat ...]"
        self.do_assert(not my_re.search(fr"(\[{char_desc}\]).*\1",
                                        test_script_contents, flags=my_re.DOTALL|my_re.IGNORECASE))
        # ex (see above): "# Input:\n#   Nothing to test [grinning...]\n#\n# Output:\n#   Nothing to test [Grinning...]"
        self.do_assert(my_re.search(fr"(\[{char_desc}\]).*\1", output,
                                    flags=my_re.DOTALL|my_re.IGNORECASE))

        # Make sure no emoticon byte sequences in UTF-8 sequences for output, although in script.
        # Note: Uses broader UTF8-based tests than Unicode character DB used in script.
        # Also, regex is done over byte sequences to account for misformed input.
        ## TODO3: use testing file for data not the tested script
        script_contents = system.read_file(self.script_file)
        loose_emoticon_regex = br"[\xE0-\xFF][\x80-\xFF]{1,3}"
        # ex: "# EX: convert_emoticons("天気") => "天気"   # Japanese for weather"
        self.do_assert(my_re.search(loose_emoticon_regex, script_contents.encode()))
        output_sans_cjk_examples = my_re.sub(r"^.*(Chinese|Japanese).*$", "", output,
                                             flags=my_re.MULTILINE)
        # ex: [same as above because CJK preserved]
        bad_match = my_re.search(loose_emoticon_regex, output_sans_cjk_examples.encode())
        debug.trace_expr(5, bad_match, output_sans_cjk_examples)
        self.do_assert(not bad_match)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_over_script_sans_comments(self):
        """Makes sure works as expected over script itself"""
        debug.trace(4, f"TestIt.test_over_script_sans_comments(); self={self}")

        # Strip comments from script and run conversion over it
        script_contents = system.read_file(__file__)
        script_contents_sans_comments = my_re.sub("#.*\n", "\n", script_contents)
        debug.trace_expr(6, script_contents_sans_comments)
        system.write_file(self.temp_file, script_contents_sans_comments)
        output = self.run_script(options="", data_file=self.temp_file)

        # There should be no extended ascii bytes
        self.do_assert(not my_re.search(b"[\x80-\xFF]", script_contents_sans_comments.encode()))
        self.do_assert(not my_re.search(b"[\x80-\xFF]", output.encode()))
        return

    @trap_exception
    def test_misc(self):
        """Test direct calls for conversion"""
        debug.trace(4, f"TestIt2.test_whatever(); self={self}")
        convert_emoticons = THE_MODULE.convert_emoticons
        cool_smile = "\U0001F60E"        # 😎
        self.do_assert(convert_emoticons(cool_smile) == "[smiling face with sunglasses]")
        self.do_assert(convert_emoticons(cool_smile, strip=True) == "")
        chinese_age = "\uF9A8"           # 令 ("age" in Chinese)
        self.do_assert(convert_emoticons(chinese_age) == chinese_age)
        return


#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Note:
# - This is work-in-progress code in much need of improvement.
# - For example, the code was adapted from type hinting checks,
#   but not sufficiently updated (see type-hinting-integration repo).
# - To disable caching of pytests, use the PYTEST_ADDOPTS environment variable externally
#   $ export PYTEST_ADDOPTS="--cache-clear"
#
# TODO0:
# - Rework the tests to be less brittle. This might be added to Github actions with
#   other code check scripts to be run once a week or so. Therefore, a single failure
#   should not cause the test suite to fail. Instead, use percentage thresholds.
#
# TODO1:
# - Remember to check for mezcla wrappers before calling functions directly (e.g., for
#   tracing and sanity checks):
#      $ para-grep getsize ~/mezcla/*.py
#   where para-grep defined in tomohara-aliases.bash of companion shell-scripts repo
# - This is good for one's pre-commit checklist (e.g., along with running python-lint).
#
# TODO2:
# - Try to make output easier to review for pytest summary.
#
# TODO3:
# - Keep most comments focused on high-level, covering the intention of the code.
#   Avoid getting into the nitty-gritty details unless it is a tricky algorithm.
#   (Moreover, tricky algorithms in general should be avoided unless critical.)
#
# TODO4:
# - Avoid numbering steps, because it becomes a maintenance issue.
# - Similarly, avoid references like "script 1", etc.
#

"""
Miscellaneous tests for mezcla_to_standard module
"""

# Standard packages
import difflib

# Installed packages
import pytest
from pydantic import BaseModel

# Local packages
from mezcla import system, debug, glue_helpers as gh, misc_utils
from mezcla.my_regex import my_re
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla.tests.common_module import RUN_SLOW_TESTS, fix_indent

# Constants, including environment variables
MEZCLA_DIR = gh.form_path(gh.dir_path(__file__), "..")
debug.assertion(system.file_exists(gh.form_path(MEZCLA_DIR, "debug.py")))
MIN_MEZCLA_SCRIPTS_COUNT = 50
LINE_IMPORT_PYDANTIC = "from pydantic import validate_call\n"
TEST_GLOB = system.getenv_value(
    "TEST_GLOB",
    None,
    description="Specify glob pattern for files to test"
)
TEST_FILES = system.getenv_value(
    "TEST_FILES",
    None,
    description="Comma-separated list of files to test"
)
debug.assertion(not (TEST_GLOB and TEST_FILES))
SKIP_WARNINGS = system.getenv_bool(
    "SKIP_WARNINGS", 
    False,
    description="Skip warning comments for tests without standard equivalents"
)
OMIT_SLOW_TESTS = not (RUN_SLOW_TESTS or TEST_GLOB or TEST_FILES)
OMIT_SLOW_REASON = "this will take a while"

## OLD: class ScriptComparison(BaseModel)
class ScriptComparisonMetrics(BaseModel):
    """Pydantic model for tracking script conversion metrics"""
    original_script: str
    converted_script: str
    total_original_lines: int
    total_converted_lines: int
    lines_added: int
    lines_removed: int
    lines_warned: int


class TestM2SBatchConversion(TestWrapper):
    """Class for batch conversion of test usage of equivalent calls for mezcla_to_standard"""

    script_module = "mezcla.mezcla_to_standard"

    def setup_test_directories(self, temp_base):
        """Setup and create required test directories"""
        dirs = {
            'typehint_org': f"{temp_base}-typehint_org",
            'typehint_m2s': f"{temp_base}-typehint_m2s",
            'mezcla_m2s': f"{temp_base}-mezcla_m2s",
            'pytest_org': f"{temp_base}-pytest_org",
            'pytest_m2s': f"{temp_base}-pytest_m2s"
        }
        for dir_path in dirs.values():
            system.create_directory(dir_path)   
        return dirs

    def process_pytest_results(self, pytest_output):
        """Extract pass/fail counts from pytest output"""
        failed = sum(map(int, my_re.findall(r"(\d+) x?failed", pytest_output)))
        passed = sum(map(int, my_re.findall(r"(\d+) x?passed", pytest_output)))
        return passed, failed
    
    def count_methods(self, code_content):
        """
        Count number of method definitions in code.
        
        Args:
            code_content: String containing the code to analyze
            
        Returns:
            Number of method definitions found
        """
        method_pattern = r"^\s*def\s+\w+\s*\(.*\):"
        # return sum(1 for _ in my_re.findall(r"^\s*def\s+\w+\s*\(.*\):", code_content, my_re.MULTILINE))
        return len(my_re.findall(method_pattern, code_content, my_re.MULTILINE))


    def validate_mez2std_conversion(self, script_path, dirs, m2s_path):
        """Process single script conversion and validation"""
        script = gh.basename(script_path)
        
        output, output_file = self.get_mezcla_command_output(
            m2s_path=m2s_path,
            script_path=script_path,
            option="to-standard",
            output_path=dirs['mezcla_m2s']
        )
        
        transformed_org = self.transform_for_validation(script_path)
        script_path_org = gh.form_path(dirs['typehint_org'], script)
        system.write_file(filename=script_path_org, text=transformed_org)
        
        transformed_m2s = self.transform_for_validation(output_file)
        script_path_m2s = gh.form_path(dirs['typehint_m2s'], script)
        system.write_file(filename=script_path_m2s, text=transformed_m2s)
        
        return {
            'script': script,
            'method_count': self.count_methods(output),
            'org_path': script_path_org,
            'm2s_path': script_path_m2s
        }
    
    def calculate_failure_metrics(self, validation_results, num_scripts):
        """Calculate and validate failure metrics"""
        fail_count = 0
        for result in validation_results:
            total_count = sum(result['m2s_passed'])
            if result['m2s_passed'][1] > result['org_passed'][1]:
                fail_count += 1
            bad_pct = round(fail_count * 100 / total_count, 2) if total_count else 0
            debug.assertion(bad_pct < 20)
            
        overall_bad_pct = round(fail_count * 100 / num_scripts, 2) if num_scripts else 0
        return overall_bad_pct
    
    # TEMPORARILY COMMENTED
    def run_validation_tests(self, script_info, dirs):
        """Run and compare validation tests for original and converted code"""
        # Create and run tests for original
        test_file_org = self.create_test_file(script_info['org_path'], dirs['typehint_org'])
        pytest_result_org = gh.run(f"PYTHONPATH='{dirs['typehint_org']}' pytest {test_file_org}")
        
        # Create and run tests for converted
        test_file_m2s = self.create_test_file(script_info['m2s_path'], dirs['typehint_m2s'])
        pytest_result_m2s = gh.run(f"PYTHONPATH='{dirs['typehint_m2s']}' pytest {test_file_m2s}")
        
        # Save test results
        org_result_path = gh.form_path(dirs['pytest_org'], gh.basename(script_info['script'], '.py')) + '.txt'
        m2s_result_path = gh.form_path(dirs['pytest_m2s'], gh.basename(script_info['script'], '.py')) + '.txt'
        # system.write_file(pytest_result_org, org_result_path)
        # system.write_file(pytest_result_m2s, m2s_result_path)
        system.write_file(org_result_path, pytest_result_org)
        system.write_file(m2s_result_path, pytest_result_m2s)
        
        return {
            'org_passed': self.process_pytest_results(pytest_result_org),
            'm2s_passed': self.process_pytest_results(pytest_result_m2s),
            'result_paths': (org_result_path, m2s_result_path)
        }

    def get_mezcla_scripts(self):
        """Returns list of paths for python scripts in MEZCLA_DIR.
        Note: Uses TEST_GLOB or TEST_FILES instead if defined
        """
        file_names = []
        if TEST_GLOB:
            file_names = gh.get_matching_files(TEST_GLOB)
        elif TEST_FILES:
            file_names = misc_utils.extract_string_list(TEST_FILES)
        else:
            file_names = [f for f in system.read_directory(MEZCLA_DIR) if f.endswith(".py")]
        debug.trace_expr(6, file_names)
        result = [(gh.form_path(MEZCLA_DIR, f) if not system.file_exists(f) else f)
                  for f in file_names]
        debug.trace(5, f"get_mezcla_scripts() => {result!r}")
        return result

    # Helper Script: Get the output of the execution of mezcla_to_standard.py (w/ options)
    def get_mezcla_command_output(self, m2s_path, script_path, option, skip_warnings=SKIP_WARNINGS, output_path="/dev/null"):
        """Executes the mezcla script externally (option: to_standard, metrics)"""
        warning_option = ("--skip-warnings" if skip_warnings else "")
        if output_path != "/dev/null":
            output_file = f"{output_path}/_mez2std_{gh.basename(script_path, '.py')}.py"
            command = f"python3 {m2s_path} --{option} {script_path} {warning_option} | tee {output_file}"
        else:
            output_file = output_path
            command = f"python3 {m2s_path} --{option} {script_path} {warning_option} > {output_path}"
        print("\nCommand from get_mezcla_command_output:", command)
        output = gh.run(command)
        return output, output_file

    # Helper Script: Get absolute path of "mezcla_to_standard.py"
    ## OLD_NAME: get_m2s_path
    def get_mezcla_to_standard_script_path(self):
        """Returns the path of mezcla_to_standard.py"""        
        return gh.form_path(MEZCLA_DIR, "mezcla_to_standard.py")

    # Helper Script: Use Pydantic class to find comparison in the script
    ## OLD_NAME: compare_scripts
    def helper_compare_converted_scripts(self, original_script_path: str, converted_script_path: str) -> ScriptComparisonMetrics:
        """Uses Pydantic to compare the contents between the original & converted scripts"""
        original_code = system.read_file(original_script_path)
        converted_code = system.read_file(converted_script_path)
        
        original_lines = original_code.splitlines()
        converted_lines = converted_code.splitlines()

        diff = difflib.unified_diff(original_code.splitlines(), converted_code.splitlines())
        differences = [line for line in diff if line.strip()]

        lines_added = len([line for line in differences if line.startswith("+") and not line.startswith("+++")])
        lines_removed = len([line for line in differences if line.startswith("-") and not line.startswith("---")])
        lines_warned = len([line for line in converted_lines if 'WARNING not supported' in line])

        return ScriptComparisonMetrics(
            original_script=original_script_path,
            converted_script=converted_script_path,
            differences=differences,
            total_original_lines=len(original_lines),
            total_converted_lines=len(converted_lines),
            lines_added=lines_added,
            lines_removed=lines_removed,
            lines_warned=lines_warned
        )
    
    ## OLD_NAME: calculate_score
    def helper_calculate_heuristic_score(
            self, 
            lines_original: int, 
            lines_converted: int,
            lines_added: int,
            lines_removed: int, 
            lines_warned: int,
            epsilon: float = 0.001,
            wt_size_difference: float = 0.40,
            wt_added: float = 0.25,
            wt_removed: float = 0.25,
            wt_warning: float = 0.10
        ):
        """
        Calculates a highly sensitive efficiency score based on the changes made during conversion,
        with increased emphasis on line changes and size differences.
        
        Arguments:
            lines_original: Number of lines in the original script.
            lines_converted: Number of lines in the converted script.
            lines_added: Number of lines added during conversion.
            lines_removed: Number of lines removed during conversion.
            lines_warned: Number of warnings generated.
            epsilon: A small value to prevent division by zero.
            wt_size_difference, wt_added, wt_removed, wt_warning: Weights for each metric.
            
        Returns:
            A normalized efficiency score between 0.0 and 1.0.
        """
        if lines_original == 0 or lines_converted <= 1:
            return 0.0

        size_ratio = min(1.0, abs(lines_converted - lines_original) / (lines_original + epsilon))
        line_add_ratio = min(1.0, lines_added / (lines_original + epsilon))
        line_removed_ratio = min(1.0, lines_removed / (lines_original + epsilon))
        line_warning_ratio = min(1.0, lines_warned / (lines_original + epsilon))

        # Penalize if warnings exceed 10% of original lines
        if lines_warned > lines_original * 0.1:  
            wt_warning += 0.1
            wt_size_difference -= 0.05

        weighted_sum = (
            size_ratio * wt_size_difference +
            line_add_ratio * wt_added +
            line_removed_ratio * wt_removed +
            line_warning_ratio * wt_warning
        )

        minimal_change_bonus = 0.02 if (size_ratio < 0.05 and line_warning_ratio < 0.05) else 0.0
        result = max(0.0, min(1.0, 1.0 - weighted_sum + minimal_change_bonus))
        return round(result, 4)

    def transform_for_validation(self, file_path):
        """Creates a copy of the script for validation of argument calls (using pydantic)"""
        content = system.read_file(file_path)
        content = my_re.sub(r"^def", r"@validate_call\ndef", content, flags=my_re.MULTILINE)
        content = LINE_IMPORT_PYDANTIC + content
        return content

    def create_test_function(self, module_name, function_name):
        """Creates a test function template for a given function name"""
        code = (
            f"""
            def test_{function_name}():
                from {module_name} import {function_name}
                assert callable({function_name})
                # Add appropriate function calls and assertions here
                try:
                    {function_name}()  # Example call, modify as needed
                except Exception as e:
                    assert False, f"Function {function_name} raised an exception: {{e}}"
            """
            )
        result = fix_indent(code)
        debug.trace(6, f"create_test_function({module_name}, {function_name}) => {result!r}")
        return result

    def create_test_file(self, script_path, test_dir):
        """Creates a test file for a given script"""
        script_name = gh.basename(script_path, ".py")
        function_names = self.extract_function_names(script_path)
        test_file_content = "\n".join([self.create_test_function(script_name, fn) for fn in function_names])
        test_file_dir = gh.form_path(test_dir, "tests")
        system.create_directory(test_file_dir)
        test_file_path = gh.form_path(test_file_dir, f"test_{script_name}.py")
        system.write_file(filename=test_file_path, text=test_file_content)
        return test_file_path

    def extract_function_names(self, file_path):
        """Extracts function names from a script"""
        content = system.read_file(file_path)
        return my_re.findall(r"^def (\w+)", content, flags=my_re.MULTILINE)

    ## EXPERIMENTAL: check_mezcla_wrappers in converted
    def check_mez2std_wrappers(self, converted_code: str) -> dict:
        """Check for proper use of mezcla wrappers in converted code"""
        wrapper_stats = {
            'direct_calls': [],
            'wrapper_calls': []
        }
        
        wrapper_patterns = {
            'getsize': r'(?:gh\.file_size|os\.path\.getsize)',
        }
        
        for func, *_ in wrapper_patterns.items():
            direct_calls = my_re.findall(rf'(?<!gh\.){func}\(', converted_code)
            wrapper_calls = my_re.findall(rf'gh\.{func}\(', converted_code)

            if direct_calls:
                wrapper_stats['direct_calls'].extend(direct_calls)
            if wrapper_calls:
                wrapper_stats['wrapper_calls'].extend(wrapper_calls)
                
        return wrapper_stats

    ## TOFIX: TestM2SBatchConversion.test_check_mez2std_wrappers() missing 2 required positional arguments: 'converted_code' and 'expected'
    @pytest.mark.skipif(OMIT_SLOW_TESTS, reason=OMIT_SLOW_REASON)
    @pytest.mark.parametrize("converted_code, expected", [
        (
            """
            import os
            def analyze_file(file_path):
                size = os.path.getsize(file_path)
                return size
            def get_wrapper_size(file_path):
                size = gh.file_size(file_path)
                return size
            """,
            {
                'direct_calls': ['getsize('],
                'wrapper_calls': ['file_size(']
            }
        ),
        (
            """
            def another_function():
                return gh.file_size('/path/to/file')
            """,
            {
                'direct_calls': [],
                'wrapper_calls': ['file_size(']
            }
        ),
        (
            """
            def no_calls_here():
                return "Nothing interesting"
            """,
            {
                'direct_calls': [],
                'wrapper_calls': []
            }
        )
    ])
    def test_check_mez2std_wrappers(self, converted_code, expected):
        """Test for check_mez2std_wrappers function in the test script"""
        result = self.check_mez2std_wrappers(converted_code)
        assert result == expected, f"Failed for input:\n{converted_code}"
    
    @pytest.mark.xfail
    @pytest.mark.skipif(OMIT_SLOW_TESTS, reason=OMIT_SLOW_REASON)
    def test_mez2std_heuristic(self):
        """
        Tests the conversion quality of mezcla scripts using various heuristic metrics.
        
        This test evaluates script conversion efficiency by:
        - Processing multiple scripts
        - Checking individual script efficiency
        - Tracking overall conversion success
        """
        
        scripts = self.get_mezcla_scripts()
        if not scripts:
            pytest.skip("No scripts found to test")
        
        m2s_path = self.get_mezcla_to_standard_script_path()
        output_dir = gh.get_temp_dir()
        
        # Initialize metrics tracking
        metrics = {
            'processed': 0,
            'successful': 0,
            'failed': 0,
            'efficiency_scores': [],
            'clean_conversions': 0,
            'warned_conversions': 0,
            'failed_scripts': []
        }
        
        print("\nProcessing scripts for heuristic evaluation:")
        
        for script_path in scripts:
            script_name = gh.basename(script_path)
            metrics['processed'] += 1
            print(f"\nScript {metrics['processed']}/{len(scripts)}: {script_name}")
            
            try:
                # Execute mezcla command and generate output
                output, output_file = self.get_mezcla_command_output(
                    m2s_path=m2s_path,
                    script_path=script_path,
                    option="to-standard",
                    output_path=output_dir
                )
                
                # Validate output file generation
                if not gh.file_exists(output_file):
                    raise RuntimeError("Conversion failed - no output file generated")
                
                # Track warning status
                if "WARNING" in output:
                    metrics['warned_conversions'] += 1
                else:
                    metrics['clean_conversions'] += 1
                
                # Compare original and converted scripts
                comparison = self.helper_compare_converted_scripts(script_path, output_file)
                
                # Calculate efficiency score
                efficiency = self.helper_calculate_heuristic_score(
                    lines_original=comparison.total_original_lines,
                    lines_converted=comparison.total_converted_lines,
                    lines_added=comparison.lines_added,
                    lines_removed=comparison.lines_removed,
                    lines_warned=comparison.lines_warned
                )
                
                # Printing script metrics
                print(
                    "\nScript Metrics (Lines):",
                    "\nOriginal:", comparison.total_original_lines,
                    "\nConverted:", comparison.total_converted_lines,
                    "\nAdded:", comparison.lines_added,
                    "\nRemoved:", comparison.lines_removed,
                    "\nWarned:", comparison.lines_warned
                )


                # Assert individual script efficiency
                assert efficiency >= 0.75, (
                    f"Script {script_name} efficiency ({efficiency:.2f}) "
                    "below minimum threshold (0.75)"
                )
                
                metrics['efficiency_scores'].append(efficiency)
                metrics['successful'] += 1
                
                # Print script processing details
                print(f"✓ Processed successfully (Efficiency: {efficiency:.2f})")

            except Exception as e:
                # Track and report failed scripts
                metrics['failed'] += 1
                metrics['failed_scripts'].append((script_name, str(e)))
                print(f"✗ Failed: {str(e)}")
        
        # Print test summary
        self.print_test_summary(metrics)

        # Calculate and validate overall test performance
        success_rate = (metrics['successful'] / metrics['processed']) * 100
        avg_efficiency = (
            sum(metrics['efficiency_scores']) / metrics["processed"] 
            if metrics['efficiency_scores'] else 0
        )

            
        assert success_rate >= 80, (
            f"Success rate ({success_rate:.1f}%) below minimum threshold (80%)"
        )
        assert avg_efficiency >= 0.75, (
            f"Average efficiency ({avg_efficiency:.2f}) below minimum threshold (0.75)"
        )
    
    # Check that 50+ scripts are collected
    def test_get_mezcla_scripts(self):
        """Tests if all Python3 scripts are in MEZCLA_DIR"""
        min_mezcla_scripts_count = (MIN_MEZCLA_SCRIPTS_COUNT
                                    if not (TEST_GLOB or TEST_FILES) else 1)
        assert len(self.get_mezcla_scripts()) >= min_mezcla_scripts_count    

    ## OLD_NAME: test_mezcla_scripts_batch_conversion(self):
    @pytest.mark.xfail
    @pytest.mark.skipif(OMIT_SLOW_TESTS, reason=OMIT_SLOW_REASON)
    def test_mez2std_batch_conversion(self):
        """Test for batch conversion of mezcla scripts to standard scripts."""
        print("\n=== Starting Batch Conversion Test (mez2std) ===\n")
        scripts = self.get_mezcla_scripts()
        if not scripts:
            print("[ERROR] No mezcla scripts found for conversion.")
            return
        
        print(f"[DEBUG] Found {len(scripts)} mezcla scripts to process.")
        
        m2s_path = self.get_mezcla_to_standard_script_path()
        if not m2s_path:
            print("[ERROR] Conversion script path is not set.")
            return

        print(f"[DEBUG] Conversion script path: {m2s_path}")
        
        output_path = gh.get_temp_dir()
        if not output_path:
            print("[ERROR] Temporary directory for output is not available.")
            return

        print(f"[DEBUG] Output directory: {output_path}\n")

        ## NEW: Added stats for tracking successful conversions 
        successful_conversions = 0
        total_scripts = len(scripts)

        for idx, script_path in enumerate(scripts, start=1):
            script_name = gh.basename(script_path)
            print(f"\n# [DEBUG {idx}/{len(scripts)}] Processing script: {script_name}")

            # Attempt to run the conversion command
            try:
                ## OLD: Unused variable 'output'
                # output, output_file = self.get_mezcla_command_output(
                #     m2s_path=m2s_path, 
                #     script_path=script_path, 
                #     option="to-standard", 
                #     output_path=output_path
                # )
                output_file = self.get_mezcla_command_output(
                    m2s_path=m2s_path, 
                    script_path=script_path, 
                    option="to-standard", 
                    output_path=output_path
                )[1]

                print(f"[DEBUG] Command output file: {output_file}")
            except Exception as e:
                print(f"[ERROR] Failed to run conversion command for {script_name}: {e}")
                continue

            # Check if the output file was created
            if not gh.file_exists(output_file):
                print(f"[ERROR] Output file not generated for {script_name}.")
                continue

            try:
                original_size = gh.file_size(script_path)
                converted_size = gh.file_size(output_file)
                print(f"[DEBUG] Original size: {original_size} bytes, Converted size: {converted_size} bytes")

                assert 0.8 * original_size <= converted_size <= 1.2 * original_size, (
                    f"\n[ERROR] Converted size {converted_size} bytes is not within 20% of the original size {original_size} bytes"
                )
            except Exception as e:
                print(f"[ERROR] File size comparison failed for {script_name}: {e}")
                continue

             # Check content integrity
            try:
                converted_content = system.read_file(output_file)
                if not converted_content.strip():
                    print(f"[ERROR] Converted file {output_file} is empty.")
                else:
                    line_count = len(converted_content.splitlines())
                    print(f"[DEBUG] Converted file {output_file} contains {line_count} lines.")
                    successful_conversions += 1
            except Exception as e:
                print(f"[ERROR] Failed to read converted file {output_file}: {e}")
                continue

        # Validate the output directory contents
        try:
            converted_files = gh.get_matching_files(gh.form_path(output_path, "*.py"))
            print(f"[DEBUG] Converted files in output directory: {len(converted_files)}")
            assert len(converted_files) == len(scripts), (
                f"[ERROR] Mismatch in expected ({len(scripts)}) and actual converted files ({len(converted_files)})."
            )
        except Exception as e:
            print(f"[ERROR] Final directory verification failed: {e}")

        print("\n=== Batch Conversion Test Completed ===\n")

        # Assert that at least 80% of scripts were converted successfully
        success_percentage = (successful_conversions / total_scripts) * 100
        assert success_percentage >= 80, (
            f"[ERROR] Only {success_percentage:.2f}% of scripts converted successfully. "
            f"Expected at least 80%. Successful conversions: {successful_conversions}/{total_scripts}"
        )
        print(f"[SUCCESS]: {successful_conversions} out of {total_scripts} passed (Efficiency: {success_percentage}% >= 80%)")


    ## COMMENTED: Activities performed by test_mez2std_heuristic
    ## NOTE: Could be seen as an alternative to calculate detailed efficiency
    ## OLD_NAME: def test_mezcla_scripts_metrics
    # @pytest.mark.xfail
    # @pytest.mark.skipif(OMIT_SLOW_TESTS, reason=OMIT_SLOW_REASON)   
    # def test_mez2std_conversion_efficiency(self, threshold=25):
    #     """Tests external scripts through mezcla using metrics option (TODO: Write better description)
    #     Note: Provides alternative "conversion efficiency" to test_mezcla_scripts_compare
    #     """
    #     ## TODO2: better motivate the use of the "efficiency" metric and use a better threshold
    #     debug.trace(6, f"test_exteral_scripts({self})")
        
    #     print(f"\nEfficiency Scores (out of 100 / threshold={threshold}):\n")
    #     # Run conversion
    #     scripts = self.get_mezcla_scripts()[0]
    #     m2s_path = self.get_m2s_path()
    #     option = "metrics"
    #     output_path = "/dev/null"

    #     for idx, script_path in enumerate(scripts, start=1):
    #         script = gh.basename(script_path)
    #         output, _output_file = self.get_mezcla_command_output(m2s_path, script_path, option, output_path)

    #         # Use regex to search calls replaced and warnings added
    #         calls_replaced = my_re.search(r"Calls replaced:\t(\d+)", output)
    #         warnings_added = my_re.search(r"Warnings added:\t(\d+)", output)
    #         calls_replaced_num = int(calls_replaced.group(1)) if calls_replaced else None
    #         warnings_added_num = int(warnings_added.group(1)) if warnings_added else None

    #         efficiency = (
    #             round((calls_replaced_num * 100) / (calls_replaced_num + warnings_added_num), 2)
    #             if calls_replaced_num != warnings_added_num
    #             else 0
    #         )
    #         print(f"#{idx} {script}: {efficiency}")
    #         assert efficiency >= threshold

    ## NEW: Alternative pytest metrics to check conversion efficiency
    @pytest.mark.xfail
    @pytest.mark.skipif(OMIT_SLOW_TESTS, reason=OMIT_SLOW_REASON)
    def test_mez2std_conversion_efficiency(self, efficiency_threshold=0.7):
        """
        Evaluates conversion efficiency using a multi-factor scoring.
        
        Complexity Score Calculation:
        - Considers calls replaced vs warnings added
        - Penalizes scripts with high warning-to-replacement ratio
        - Provides overall conversion complexity assessment
        """
        print("\n=== Starting Mezcla Conversion Complexity Analysis ===\n")

        
        scripts = self.get_mezcla_scripts()
        if not scripts:
            print("[ERROR] No mezcla scripts found for analysis.")
            return
        
        m2s_path = self.get_mezcla_to_standard_script_path()
        if not m2s_path:
            print("[ERROR] Conversion script path is not set.")
            return

        total_scripts = len(scripts)
        complexity_scores = []

        for idx, script_path in enumerate(scripts, start=1):
            script_name = gh.basename(script_path)
            print(f"\n[ANALYSIS {idx}/{total_scripts}] Processing script: {script_name}")

            try:
                output, *_ = self.get_mezcla_command_output(
                    m2s_path=m2s_path, 
                    script_path=script_path, 
                    option="metrics", 
                    output_path="/dev/null"
                )

                # Extract metrics using regex
                calls_replaced = my_re.search(r"Calls replaced:\t(\d+)", output)
                warnings_added = my_re.search(r"Warnings added:\t(\d+)", output)
                
                calls_replaced_num = int(calls_replaced.group(1)) if calls_replaced else 0
                warnings_added_num = int(warnings_added.group(1)) if warnings_added else 0

                # Calculate complexity score
                total_modifications = calls_replaced_num + warnings_added_num
                if total_modifications > 0:
                    complexity_score = (
                        calls_replaced_num / total_modifications
                    )
                else:
                    complexity_score = 0

                complexity_scores.append(complexity_score)
                
                print(f"[DEBUG] {script_name}: Calls Replaced = {calls_replaced_num}, "
                      f"Warnings Added = {warnings_added_num}, "
                      f"Complexity Score = {complexity_score:.2f}")

            except Exception as e:
                print(f"[ERROR] Analysis failed for {script_name}: {e}")
                complexity_scores.append(0)

        # Calculate overall metrics
        avg_complexity = sum(complexity_scores) / total_scripts if complexity_scores else 0
        
        print("\n=== Conversion Complexity Analysis Summary ===")
        print(f"Total Scripts Analyzed: {total_scripts}")
        print(f"Average Complexity Score: {avg_complexity:.2f}")
        print(f"Complexity Threshold Score: {efficiency_threshold:.2f}")

        # Assert that the average complexity meets the threshold
        assert avg_complexity >= efficiency_threshold, (
            f"[ERROR] Average complexity score {avg_complexity:.2f} "
            f"is below the required threshold of {efficiency_threshold}"
        )

    # NEW: Test summary print handled by a different method
    def print_test_summary(self, metrics: dict):
        """Prints a detailed summary of test results based on the metrics."""
        print("\n=== Heuristic Test Summary ===")
        print(f"Total Scripts Processed: {metrics['processed']}")
        
        print("\nConversion Statistics:")
        print(f"- Successful Conversions: {metrics['successful']}")
        print(f"- Failed Conversions: {metrics['failed']}")
        success_rate = (metrics['successful'] / metrics['processed']) * 100 if metrics['processed'] > 0 else 0
        print(f"- Success Rate: {success_rate:.1f}%")
        
        print("\nQuality Metrics:")
        print(f"- Clean Conversions: {metrics['clean_conversions']}")
        print(f"- Conversions with Warnings: {metrics['warned_conversions']}")
        avg_efficiency = (sum(metrics['efficiency_scores']) / metrics["processed"]
                        if metrics['efficiency_scores'] else 0)
        print(f"- Average Efficiency Score: {avg_efficiency:.2f}")
        
        if metrics['failed_scripts']:
            print("\nFailed Scripts:")
            for script, error in metrics['failed_scripts']:
                print(f"- {script}: {error}")
        
        print()

if __name__ == "__main__":
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Tests for format_profile module
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_train_language_model.py 

## TODO1: carefully review test/template.py
## TODO2: use a helper method to cut down on redundant tests; for example,
##    def check_format_profile_key(self, key):
##        ... run_script(env_options=f"PROFILE_KEY={key}", ...)
##    def test_formatprofile_PK_time(self): self.check_format_profile_key("time")
## NOTE:
## - Only need to test a few main keys in any depth, so generalize test_formatprofile_PK_calls
## so that it can handle alternative sort orders without substantial revision.
## - The minor keys can just check for "Ordered by" label differences.


"""Tests for format_profile module"""

# Standard packages
## OLD: import re

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
## OLD: from mezcla.my_regex import my_re
## OLD: from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
## OLD: from mezcla.unittest_wrapper import trap_exception

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
import mezcla.format_profile as THE_MODULE

class TestFormatProfile(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    use_temp_base_dir = True    # treat TEMP_BASE as directory
    testing_script = gh.resolve_path("test_glue_helpers.py", heuristic=True)
    
    # For testing, let the dir be ./mezcla/mezcla/tests
    # TODO: Find a way to denote a random number in SAMPLE OUTPUT which may differ in each execution

    # ## OLD: Doesn't use helper function
    # @pytest.mark.xfail                   # TODO: remove xfail
    # @trap_exception                      # TODO: remove when debugged
    # def test_formatprofile_PK_calls(self):
    #     "Ensures that test_formatprofile_PK_calls works as expected"
    #     ## TODO3: reword commment because test_formatprofile_PK_calls is not testing itself;
    #     ## Instead, use something like "ensure that PROFILE_KEY=calls works as expected".

    #     key_arg = "calls"
    #     ## OLD: testing_script = "test_glue_helpers.py"
    #     testing_script = gh.resolve_path("simple_main_example.py")
    #     ## OLD: SAMPLE_OUTPUT = ["1    0.001    0.000    0.000    0.000 cacheprovider.py:307(pytest_report_collectionfinish)", "{method 'popleft' of 'collections.deque' objects}"]

    #     # Sample output:
    #     #     Ordered by: call count
    #     # 
    #     #     ncalls  tottime  percall  cumtime  percall filename:lineno(function)
    #     #     3323    0.000    0.000    0.000    0.000 {built-in method builtins.isinstance}
    #     #     2670    0.000    0.000    0.000    0.000 {method 'rstrip' of 'str' objects}
    #     #     1825/1723    0.000    0.000    0.000    0.000 {built-in method builtins.len}
    #     #     1443    0.000    0.000    0.000    0.000 {method 'append' of 'list' objects}
    #     #
        
    #     debug.trace(4, f"test_formatprofile_PK_calls(); self={self}")
    #     ## OLD: empty_file1 = self.get_temp_file()
    #     ## OLD: profile_log  = self.get_temp_file()
    #     profile_data = self.temp_file + "-profile.data"
    #     ## OLD: test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
    #     # note: runs simple_main_example.py over itself; for example,
    #     #   python -m cProfile -o /tmp/tmpp6zg9a5o/test-1-profile.data simple_main_example.py simple_main_example.py
    #     test_command_1 = f"python -m cProfile -o {profile_data} {testing_script} {testing_script}"
    #     ## BAD: test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"

    #     gh.run(test_command_1)
    #     ## BAD: gh.run(test_command_2)

    #     ## BAD: output = gh.read_file(empty_file1)
    #     output = self.run_script(env_options=f"PROFILE_KEY={key_arg}", data_file=profile_data)
    #     self.do_assert("Ordered by: call count" in output)
    #     my_re.search(r"^\s*(\S+)\s+.*method 'rstrip'", output, flags=my_re.MULTILINE)
    #     rstrip_count = system.to_float(my_re.group(1))
    #     my_re.search(r"^\s*(\S+)\s+.*method 'append'", output, flags=my_re.MULTILINE)
    #     join_count = system.to_float(my_re.group(1))
    #     self.do_assert(rstrip_count > join_count)

    #     return
    
    def helper_format_profile(
        self,
        key_arg,
        testing_script,
        ):
        """Helper function for format_profile tests"""
        debug.trace(4, f"helper_format_profile(); self={self}")
        
        # # OLD: Uses gh.run() approach
        # empty_file = self.get_temp_file()
        # profile_log = self.get_temp_file()
        # command1 = f"python3 -m cProfile -o {profile_log} {testing_script}"
        # command2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file}"
        # gh.run(command1)
        # gh.run(command2)
        # output = gh.read_file(empty_file)
        # return output

        ## OLD: profile_log = self.get_temp_file()
        profile_log = self.get_temp_file()
        command_cprofile = f"python3 -m cProfile -o {profile_log} {testing_script}"
        _cprofile_output = gh.run(command_cprofile)
        output = self.run_script(
            env_options=f"PROFILE_KEYS={key_arg}",
            data_file=profile_log,
        )
        assert("Ordered by:" in output)
        return output

    @pytest.mark.xfail                   # TODO: remove xfail
    # @trap_exception                      # TODO: remove when debugged
    def test_formatprofile_PK_calls(self):
        """Ensures that PROFILE_KEY=calls works as expected"""

        key_arg = "calls"
        ## OLD: testing_script = "test_glue_helpers.py"
        ## OLD: SAMPLE_OUTPUT = ["1    0.001    0.000    0.000    0.000 cacheprovider.py:307(pytest_report_collectionfinish)", "{method 'popleft' of 'collections.deque' objects}"]
        SAMPLE_OUTPUT = [
            "test_glue_helpers.py:232(test_heuristic_resolve_paths)",
            "1    0.000    0.000    0.000    0.000 tempfile.py:800(SpooledTemporaryFile)", 
        ]
        debug.trace(4, f"test_formatprofile_PK_calls(); self={self}")
        output = self.helper_format_profile(key_arg, self.testing_script)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return
        
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_cumulative(self):
        """Ensures that PROFILE_KEY=cumulative works as expected"""

        key_arg = "cumulative"
        ## OLD: testing_script = "test_glue_helpers.py"
        # OLD: SAMPLE_OUTPUT = ["<frozen importlib._bootstrap>:211(_call_with_frames_removed)q", "2    0.000    0.000    0.000    0.000 logging.py:128(_get_auto_indent)"]
        SAMPLE_OUTPUT = [
            "<frozen importlib._bootstrap_externals>:877(exec_module)", 
            "1    0.000    0.000    0.000    0.000 {built-in method posix.readlink}"
        ]

        debug.trace(4, f"test_formatprofile_PK_cumulative(); self={self}")
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        output = self.helper_format_profile(key_arg, self.testing_script)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return
    
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_cumtime(self):
        """Ensures that PROFILE_KEY=cumtime works as expected"""
        ## TODO: Find other input sample
        debug.trace(4, f"test_formatprofile_PK_cumtime(); self={self}")
        key_arg = "cumtime"
        ## OLD: testing_script = "test_glue_helpers.py"
        SAMPLE_OUTPUT = [
            "test_glue_helper.py:1(<module>)", # Incorrect Line 
            "1    0.000    0.000    0.000    0.000 {method 'fileno' of '_io.BufferedWriter' objects}"
            ]
        ## OLD
        # SAMPLE_OUTPUT = [
        #     "_hooks.py:244(__call__)qq", 
        #     "1    0.000    0.000    0.000    0.000 <attrs generated eq attr.validators._NumberValidator>:1(<module>)"
        #     ]
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"

        # gh.run(test_command_1)
        # gh.run(test_command_2)

        # output = gh.read_file(empty_file1)
        output = self.helper_format_profile(key_arg, self.testing_script)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        # return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_file(self):
        """Ensures that PROFILE_KEY=file works as expected"""

        key_arg = "file"
        ## OLD: testing_script = "test_glue_helpers.py"
        SAMPLE_OUTPUT = [
            "<frozen importlib._bootstraps>:391(cached)", 
            "1    0.000    0.000    0.000    0.000 {method 'union' of 'frozenset' objects}"
            ]
        
        debug.trace(4, f"test_formatprofile_PK_file(); self={self}")
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"

        # gh.run(test_command_1)
        # gh.run(test_command_2)
        output = self.helper_format_profile(key_arg, self.testing_script)

        # output = gh.read_file(empty_file1)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        # return
    
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_filename(self):
        """Ensures that PROFILE_KEY=filename works as expected"""

        key_arg = "filename"
        ## OLD: testing_script = "test_glue_helpers.py"
        SAMPLE_OUTPUT = [
            "1    0.000    0.000    0.000    0.000 :1(ReprEntryNativeAttributes)", 
            "6768    0.000    0.000    0.000    0.000 :1(ReprEntryAttributes)"
        ]


        ## OLD
        debug.trace(4, f"test_formatprofile_PK_filename(); self={self}")
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"

        # gh.run(test_command_1)
        # gh.run(test_command_2)

        # output = gh.read_file(empty_file1)

        output = self.helper_format_profile(key_arg, self.testing_script)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] not in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_module(self):
        """Ensures that PROFILE_KEY=module works as expected"""

        key_arg = "module"
        ## OLD: testing_script = "test_glue_helpers.py"
        # OLD: SAMPLE_OUTPUT = ["1    0.000    0.000    0.000    0.000 :1(ExceptionChainReprAttributes)"]
        SAMPLE_OUTPUT = [
            "ElementTree.pytest:1771(C14NWriterTarget)",
            "1    0.000    0.000    0.000    0.000 terminal.py:1306(_build_normal_summary_stats_line)"
        ]

        debug.trace(4, f"test_formatprofile_PK_module(); self={self}")
        output = self.helper_format_profile(key_arg, self.testing_script)
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        # return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_ncalls(self):
        """Ensures that PROFILE_KEY=ncalls works as expected"""
        ## TODO: Find other input sample

        key_arg = "ncalls"
        ## OLD: testing_script = "test_glue_helpers.py"
        SAMPLE_OUTPUT = [
            "{method 'extend' of 'collections.deque' objections}", 
            "1    0.000    0.000    0.000    0.000 test_glue_helpers.py:385(test_get_files_matching_specs)"
            ]

        debug.trace(4, f"test_formatprofile_PK_ncalls(); self={self}")

        ## OLD
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        output = self.helper_format_profile(key_arg, self.testing_script)
        # print (output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_pcalls(self):
        """Ensures that PROFILE_KEY=pcalls works as expected"""

        key_arg = "pcalls"
        ## OLD: testing_script = "test_glue_helpers.py"
        SAMPLE_OUTPUT = [
            "unix_events.py:1022(SafestChildWatcher)",
            "1    0.000    0.000    0.000    0.000 <frozen importlib._bootstrap>:294(_module_repr)"
        ]

        debug.trace(4, f"test_formatprofile_PK_pcalls(); self={self}")
        output = self.helper_format_profile(key_arg, self.testing_script)
        # SAMPLE_OUTPUT = ["{method 'extend' of 'collections.deque' objects}", "typing.py:321(__hash__)"]
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        # print(output)
        
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        # return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_line(self):
        """Ensures that PROFILE_KEY=line works as expected"""

        key_arg = "line"
        ## OLD: testing_script = "test_glue_helpers.py"
        SAMPLE_OUTPUT = [
            "{method 'with_traceback' of 'BaseExceptions' objects}", 
            "1    0.000    0.000    0.000    0.000 {method 'rjust' of 'str' objects}"
        ]
        debug.trace(4, f"test_formatprofile_PK_line(); self={self}")

        output = self.helper_format_profile(key_arg, self.testing_script)

        ## OLD
        # SAMPLE_OUTPUT = ["0.000    0.000    0.000    0.000 {method 'with_traceback' of 'BaseException' objects}", "0.000    0.000    0.000    0.000 {method 'replace' of 'kode' objects}"]
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_name(self):
        """Ensures that PROFILE_KEY=name works as expected"""

        key_arg = "name"
        ## OLD: testing_script = "test_glue_helpers.py"
        SAMPLE_OUTPUT = [
            "{method 'fileno' of '_io.BufferedReaders' objects}", 
            "6    0.000    0.000    0.000    0.000 {method 'pop' of 'collections.deque' objects}"
        ]

        debug.trace(4, f"test_formatprofile_PK_name(); self={self}")
        output = self.helper_format_profile(key_arg, self.testing_script)
        # SAMPLE_OUTPUT = ["{built-in method _csv.reader}", "{built-in method _sre.compiler}"]
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_nfl(self):
        """Ensures that PROFILE_KEY=nfl works as expected"""

        key_arg = "nfl"
        ## OLD: testing_script = "test_glue_helpers.py"
        
        SAMPLE_OUTPUT = [
            "{built-in methods _imp.is_frozen}",
            "1    0.000    0.000    0.000    0.000 {built-in method _stat.S_IMODE}", 
        ]

        debug.trace(4, f"test_formatprofile_PK_nfl(); self={self}")
        
        # SAMPLE_OUTPUT = ["{built-in method _elementtree._set_factory}", "{built-in method _imp.is_frozen}"]
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        output = self.helper_format_profile(key_arg, self.testing_script)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        # return

    ## BAD
    ## @pytest.mark.xfail                   # TODO: remove xfail
    ## def test_formatprofile_PK_name(self):
    ##     "Ensures that test_formatprofile_PK_name works as expected"

    ##     key_arg = "name"
    ##     testing_script = "test_glue_helpers.py"
    ##     SAMPLE_OUTPUT = ["{built-in method _csv.reader}", "{built-in method _sre.compiler}"]

    ##     debug.trace(4, f"test_formatprofile_PK_name(); self={self}")
    ##     empty_file1 = self.get_temp_file()
    ##     profile_log  = self.get_temp_file()
    ##     test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
    ##     test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"

    ##     gh.run(test_command_1)
    ##     gh.run(test_command_2)

    ##     output = gh.read_file(empty_file1)
    ##     assert (SAMPLE_OUTPUT[0] in output and SAMPLE_OUTPUT[1] not in output)
    ##     return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_stdname(self):
        """Ensures that PROFILE_KEY=stdname works as expected"""

        key_arg = "stdname"
        ## OLD: testing_script = "test_glue_helpers.py"

        SAMPLE_OUTPUT = [
            "zipperfile.py:1(<module>)", 
            "1    0.000    0.000    0.000    0.000 _synchronization.py:70(SemaphoreStatistics)"
        ]

        debug.trace(4, f"test_formatprofile_PK_stdname(); self={self}")

        ## OLD
        # SAMPLE_OUTPUT = [
        #     "1(ExceptionChainReprAttributes)", 
        #     "0.000    0.000    0.000    0.000 {method 'sub' of 're.Pattern' objects}"
        # ]
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)

        output = self.helper_format_profile(key_arg, self.testing_script)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_time(self):
        """Ensures that PROFILE_KEY=time works as expected"""

        key_arg = "time"
        ## OLD: testing_script = "test_glue_helpers.py"

        SAMPLE_OUTPUT = [
            "<frozen importlibrary._bootstrap_external>:380(cache_from_source)", 
            "1    0.000    0.000    0.000    0.000 glue_helpers.py:731(delete_existing_file)"
        ]

        debug.trace(4, f"test_formatprofile_PK_time(); self={self}")
        output = self.helper_format_profile(key_arg, self.testing_script)
        
        ## OLD
        # SAMPLE_OUTPUT = [
        #     "minidom.py:966(ProcessingInstruction)", 
        #     "2    0.000    0.000    0.000    0.000 _synchronization.py:24(CapacityLimiterStatistics)"
        # ]
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_formatprofile_PK_tottime(self):
        """Ensures that PROFILE_KEY=tottime works as expected"""
        ## TODO: Find other input sample
        
        key_arg = "tottime"
        ## OLD: testing_script = "test_glue_helpers.py"
        SAMPLE_OUTPUT = [
            "{method 'split' of 're.Pattern' objections}", 
            "1    0.000    0.000    0.000    0.000 cacheprovider.py:390(pytest_sessionfinish)"
        ]

        debug.trace(4, f"test_formatprofile_PK_tottime(); self={self}")
        
        ## OLD
        # empty_file1 = self.get_temp_file()
        # profile_log  = self.get_temp_file()
        # test_command_1 = f"python -m cProfile -o {profile_log} {testing_script}"
        # test_command_2 = f"PROFILE_KEY={key_arg} ../format_profile.py {profile_log} > {empty_file1}"
        # gh.run(test_command_1)
        # gh.run(test_command_2)
        # output = gh.read_file(empty_file1)
        output = self.helper_format_profile(key_arg, self.testing_script)
        # print(output)
        assert (SAMPLE_OUTPUT[0] not in output and SAMPLE_OUTPUT[1] in output)
        return

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../rgb_color_name.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_rgb_color_name.py
#
# TODO2:
# - Bite the bullet and drop the xfail's.
# - Likewise remove long-in-place xfail's in other test files (unless brittle).
#

"""Tests for rgb_color_name module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.rgb_color_name as THE_MODULE

class TestRgbColorName(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def helper_rgb_color_name(self, cmd_option:str, file_content:str):
        """Runs script over FILE_CONTENT using CMD_OPTION"""
        data_file =self.create_temp_file(contents=file_content)
        output = self.run_script(options=cmd_option, data_file=data_file)
        return output

    @pytest.mark.xfail
    def test_data_file(self):
        """Makes sure colors annotated as expected"""
        debug.trace(4, "TestRgbColorName.test_data_file()")

        content = (
            'Extracted colors:\n'
            '(255, 0, 0):  72.98% (1888)\n'
            '(0, 255, 0):  24.35% (630)\n'
            '(0, 0, 255):   2.67% (69)\n'
            '\n'
            'Pixels in output: 2587 of 11648\n'
        )
        system.write_file(self.temp_file, content)
        # =>
        #   <(255, 0, 0), red>:  72.98% (1888)
        #   <(0, 255, 0), lime>:  24.35% (630)
        #   <(0, 0, 255), blue>:   2.67% (69)
        output = self.run_script("", self.temp_file)
        self.do_assert(re.search(r"<\(0, 255, 0\), lime>", output))
        return
        
    @pytest.mark.xfail
    def test_rgb_regex(self):
        """Test the regex for RGB specification"""
        debug.trace(4, "test_rgb_regex()")

        ## OLD:
        # script = THE_MODULE.Script(skip_args=True)
        # self.do_assert(script.rgb_regex == r'\((0?x?[0-9A-F]+), (0?x?[0-9A-F]+), (0?x?[0-9A-F]+)\)')
        # line = "(39, 39, 39)   :  24.35% (630)"
        # script.process_line(line)
        # self.do_assert("<(39, 39, 39), darkslategray>" in self.get_stdout())
        
        ## OLD: 
        # rgb_regex_val = r'\\((\\d+),\\s*(\\d+),\\s*(\\d+)\\)'
        # option = f'--rgb-regex {rgb_regex_val}'

        option = "--rgb-regex '\\((\\d+),\\s*(\\d+),\\s*(\\d+)\\)'"
        color_tuple = "(0, 255, 0)"
        color = "lime"
        helper_output = self.helper_rgb_color_name(
            cmd_option=option,
            file_content=color_tuple
        )
        assert color in helper_output
    
    @pytest.mark.xfail
    def test_rgb_hex3(self):
        """Test the hex3 option"""
        debug.trace(4, "test_rgb_hex3()")
        option = "--hex3"
        hex3_val = "#f45"
        color = "tomato"
        helper_output = self.helper_rgb_color_name(
            cmd_option=option,
            file_content=hex3_val
        )
        assert color in helper_output

        ## OLD:
        # data_file =self.create_temp_file(contents=hex3_val)
        # output = self.run_script(
        #     data_file=data_file,
        #     options=option
        # )
        # assert color in output

    @pytest.mark.xfail
    def test_rgb_hex6(self):
        """Test the hex6 option"""
        debug.trace(4, "test_rgb_hex3()")
        option = "--hex6"
        hex6_val = "#a36651"
        color = "sienna"
        helper_output = self.helper_rgb_color_name(
            cmd_option=option,
            file_content=hex6_val
        )
        assert color in helper_output

        ## OLD:
        # data_file =self.create_temp_file(contents=hex6_val)
        # output = self.run_script(
        #     data_file=data_file,
        #     options=option
        # )
        # assert color in output
    
    @pytest.mark.xfail
    def test_rgb_show_hex(self):
        """Test the show-hex option"""
        debug.trace(4, "test_rgb_shiw_hex()")
        option = "--show-hex"
        color_tuple = "(39, 54, 251)"
        color = "royalblue"
        color_hex = "0x2736FB"
        helper_output = self.helper_rgb_color_name(
            cmd_option=option,
            file_content=color_tuple
        )
        assert color in helper_output
        assert color_hex in helper_output

        ## OLD: Without helper functions
        # data_file =self.create_temp_file(contents=color_tuple)
        # output = self.run_script(
        #     data_file=data_file,
        #     options=option
        # )
        # assert color and color_hex in output

    @pytest.mark.xfail
    def test_rgb_hex(self):
        """Test the hex option"""
        debug.trace(4, "test_rgb_hex()")
        option = "--hex"
        color_tuple = "(145, 128, 43)"
        color = "yellow"
        helper_output = self.helper_rgb_color_name(
            cmd_option=option,
            file_content=color_tuple
        )
        assert color in helper_output

        ## OLD:
        # data_file =self.create_temp_file(contents=color_tuple)
        # output = self.run_script(
        #     data_file=data_file,
        #     options=option
        # )
        # assert color in output

    @pytest.mark.xfail
    def test_rgb_skip_direct(self):
        """Test the skip-direct option"""
        debug.trace(4, "test_rgb_skip_direct()")

        ## NOTE: Variation of colorname for --hex and --skip-direct option
        ## NOTE: --skip-direct provides more precision with color names
        # ricekiller@pop-os:~/mezcla/mezcla$ python3 rgb_color_name.py --skip-direct input.txt
        # <(145, 128, 43), olivedrab>
        # ricekiller@pop-os:~/mezcla/mezcla$ python3 rgb_color_name.py --hex input.txt
        # <(145, 128, 43), yellow>

        option = "--skip-direct"
        color = "olivedrab"
        color_tuple = "(145, 128, 43)"
        helper_output = self.helper_rgb_color_name(
            cmd_option=option,
            file_content=color_tuple
        )
        assert color in helper_output

        ## OLD: WIthout helper function
        # data_file = self.create_temp_file(contents=color_tuple)
        # output = self.run_script(
        #     data_file=data_file,
        #     options=option
        # )
        # assert color in output
        
#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../filter_random.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_filter_random.py
#
# TODO:
# - Use actual data file (e.g., license).
# - Add some more tests.
#

"""Tests for filter_random module"""

# Standard packages
## NOTE: this is empty for now

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.filter_random as THE_MODULE

class TestFilterRandom(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    ## OLD: use_temp_base_dir = True            # treat TEMP_BASE as directory

    def setUp(self):
        """Per-test setup"""
        debug.trace(5, f"TestFilterRandom.setUp(); self={self}")
        # note: must do parent first (e.g., for temp file support)
        super().setUp()
        num_lines = 10
        data = [str(i) for i in range(num_lines)]
        debug.assertion(self.temp_file + ".data")
        system.write_lines(self.temp_file, data)
        return
    
    def run_data_file_test(self, ratio, data_file_path, expected_output):
        """Run script to include RATIO of DATA_FILE lines with EXPECTED_OUTPUT"""
        debug.trace(5, f"run_data_file_test({self}, {ratio}, {data_file_path}, {expected_output})")
        script_output = self.run_script(options=f"--ratio {ratio} --quiet --seed 13",
                                        data_file=data_file_path)
        actual_output = script_output.strip()
        expected_output = expected_output.strip()
        assert expected_output == actual_output
        return None

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_simple_data_file(self):
        """Makes sure simple canned data file works as expected"""
        debug.trace(4, f"TestFilterRandom.test_simple_data_file({self})")
        ## OLD: self.setUp()
        return self.run_data_file_test(0.15, self.temp_file, "6\n9\n")

    def test_filter_all(self):
        """Makes sure all lines are filtered out with ratio 0.0"""
        debug.trace(4, f"TestFilterRandom.test_filter_all({self})")
        ## OLD: self.setUp()
        return self.run_data_file_test(0.0, self.temp_file, "")

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_filter_none(self):
        """Makes sure no lines are filtered out with ratio 1.0"""
        debug.trace(4, f"TestFilterRandom.test_filter_none({self})")
        temp_file_contents = system.read_file(self.temp_file)
        ## OLD: self.setUp()
        return self.run_data_file_test(1.0, self.temp_file, temp_file_contents)

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Simple tests for debug.py, based on following:
#     https://stackoverflow.com/questions/16039463/how-to-access-the-py-test-capsys-from-inside-a-test
#
# Notes:
# - This can be run as follows:
#   $ SKIP_ATEXIT=1 PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_debug.py
# - For tests that capture standard error, see
#       https://docs.pytest.org/en/6.2.x/capture.html
# - This uses capsys fixture mentioned in above link.
#................................................................................
# TODO:
# - make sure trace_fmt traps all exceptions
#   debug.trace_fmt(1, "fu={fu}", fuu=1)
#                           ^^    ^^^
#

"""Tests for debug module"""

# Standard packages
import sys
from datetime import datetime
import inspect

# Installed packages
import pytest

# Local packages
## TODO: make sure atexit support disabled unless explcitly requested
##   import os; os.environ["SKIP_ATEXIT"] = os.environ.get("SKIP_ATEXIT", "1")
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
import mezcla.glue_helpers as gh
from mezcla.tests.common_module import fix_indent

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
import mezcla.debug as THE_MODULE # pylint: disable=reimported

# Environment options
# Note: These are just intended for internal options, not for end users.
# It also allows for enabling options in one place.
#
TEST_TBD = system.getenv_bool("TEST_TBD", False,
                              description="Test features to be designed: TBD")

#................................................................................
# Classes for testing

class Test_class: 
    """Test class for test_trace_object"""

    age: int = 25
    _debt: int = 5000
    __income: int = 75000
    
    def age_up(self) -> None:
        """Increase age"""
        self.age += 1
        
    def _has_debt(self) -> bool:          # pylint: disable=unused-private-member
        return self._debt > 0

    def __good_income(self) -> bool:      # pylint: disable=unused-private-member
        return self.__income > 50000

class Person:
    """Test class for test_trace_values"""
    
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f'Person("{self.name}")'

#................................................................................

class TestDebug(TestWrapper):
    """Class for test case definitions"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    stdout_text = None
    stderr_text = None
    expected_stdout_trace = None
    expected_stderr_trace = None

    def setup_simple_trace(self):
        """Common setup for simple tracing"""
        debug.trace(4, f"setup_common(): self={self}")
        self.stdout_text = "hello"
        self.stderr_text = "world"
        self.expected_stdout_trace = self.stdout_text + "\n"
        self.expected_stderr_trace = self.stderr_text + "\n"

    def test_set_level(self):
        """Ensure set_level works as expected"""
        debug.trace(4, f"test_set_level(): self={self}")
        THE_MODULE.set_level(5)
        assert THE_MODULE.trace_level == 5

    def test_get_level(self):
        """Ensure get_level works as expected"""
        debug.trace(4, f"test_get_level(): self={self}")
        THE_MODULE.trace_level = 5
        assert THE_MODULE.get_level() == 5

    def test_get_output_timestamps(self):
        """Ensure get_output_timestamps works as expected"""
        debug.trace(4, f"test_get_output_timestamps(): self={self}")
        THE_MODULE.output_timestamps = 'some-test-value'
        assert THE_MODULE.get_output_timestamps() == 'some-test-value'

    def test_set_output_timestamps(self):
        """Ensure set_output_timestamps works as expected"""
        debug.trace(4, f"test_set_output_timestamps(): self={self}")
        THE_MODULE.output_timestamps = False
        THE_MODULE.set_output_timestamps('some-test-value')
        assert THE_MODULE.output_timestamps == 'some-test-value'

    def test_trace(self):
        """Ensure trace works as expected"""
        debug.trace(4, f"test_trace(): self={self}")
        THE_MODULE.output_timestamps = True

        THE_MODULE.trace(-1, 'error foobar', indentation=' -> ')
        out,err  = self.get_stdout_stderr()
        assert " -> error foobar" in err
        assert not out

        # Test debug_file
        THE_MODULE.debug_file = sys.stdout
        THE_MODULE.trace(-1, 'some text to test debug file')
        out,err  = self.get_stdout_stderr()
        assert 'some text to test debug file' in out
        THE_MODULE.debug_file = None

        THE_MODULE.output_timestamps = False

    @pytest.mark.xfail
    def test_trace_fmtd(self):
        """Ensure trace_fmtd works as expected"""
        debug.trace(4, f"test_trace_fmtd(): self={self}")
        self.patch_trace_level(5)
        
        THE_MODULE.trace_fmtd(4, "trace_{fmtd}", **{"max_len": 5, "fmtd": "formatted"})
        stderr = self.get_stderr()
        assert "trace_forma..." in stderr
        
        self.clear_stderr()
        THE_MODULE.trace_fmtd(4, "trace_{fmtd}", **{"fmtd": "formatted"})
        stderr_2 = self.get_stderr()
        assert "trace_formatted" in stderr_2

        self.clear_stderr()
        THE_MODULE.trace_fmtd(5, "x={x}", x=1, do_whatever=True)
        stderr_3 = self.get_stderr()
        assert my_re.search(r"Unexpected keyword.*['do_whatever']", stderr_3)

    @pytest.mark.xfail
    def test_trace_object(self):
        """Ensure trace_object works as expected"""
        debug.trace(4, f"test_trace_object(): self={self}")
        obj = Test_class()
        THE_MODULE.trace_object(level=1, obj=obj, show_all=True)
        stderr = self.get_stderr()
        # <class '__main__.Test_class'> 0x7e30e9e54f10: {
        # ... _Test_class__good_income: ('<bound method Test_class.__good_income ...)
        assert my_re.search(r"Test_class[^:]+_good_income:[^:]+bound method", stderr)
        # ... _Test_class__income: 75000
        assert my_re.search(r"Test_class[^:]+_income: 75000", stderr)
        # ... _debt: 5000
        assert my_re.search(r"_debt: 5000", stderr)
        # ... _has_debt: ('<bound method Test_class._has_debt ...)
        assert my_re.search(r"_has_debt:[^:]+bound method", stderr)
        # ... age: 0
        assert my_re.search(r"age: 25", stderr)
        # ... age_up: ('<bound method Test_class.age_up of
        assert my_re.search(r"age_up:[^:]+bound method", stderr)
        
    @pytest.mark.xfail
    def test_negative_trace_object(self):
        """Negative tests for trace_object"""
        debug.trace(4, f"test_negative_trace_object(): self={self}")
        obj = Test_class()
        THE_MODULE.trace_object(level=1, obj=obj,
                                show_all=False, show_methods_etc=False, show_private=False)
        stderr = self.get_stderr()
        assert "__has_debt:" not in stderr
        ## TODO: assert "__income" not in stderr

    def test_trace_values(self):
        """Ensure trace_values works as expected"""
        debug.trace(4, f"test_trace_values(): self={self}")
        # Level -1 is used to ensure that always will run

        collection_test = [
            'foobarerror',
            'some-error',
            'another error',
        ]

        # Test normal usage
        THE_MODULE.trace_values(-1, collection_test)
        err = self.get_stderr()
        for element in collection_test:
            assert f": {element}" in err

        # Test indentation
        THE_MODULE.trace_values(-1, collection_test, indentation=' -> ')
        err = self.get_stderr()
        for i, element in enumerate(collection_test):
            assert f" -> {i}: {element}" in err

        # Test non list collection (string)
        THE_MODULE.trace_values(-1, "somevalue")
        err = self.get_stderr()
        for char in "somevalue":
            assert f": {char}" in err

        # Test non list collection (tuple)
        THE_MODULE.trace_values(-1, 123)
        err = self.get_stderr()
        assert ": 123" in err

        # Test use_repr parameter
        THE_MODULE.trace_values(-1, [Person("Kiran")], use_repr=True)
        err = self.get_stderr()
        assert 'Person("Kiran")' in err

    @pytest.mark.xfail
    def test_simple_trace_expr(self):
        """Make sure trace_expr shows 'expr1=value1; expr2=value2'"""
        debug.trace(4, f"test_trace_expr(): self={self}")
        var1 = 3
        var2 = 6
        THE_MODULE.trace_expr(debug.get_level(), var1, var2)
        err = self.get_stderr()
        assert(my_re.search(r"var1=3; var2=6;?", err))

    @pytest.mark.xfail
    def test_trace_expr_prefix(self):
        """Make sure trace_expr outputs prefix'"""
        debug.trace(4, f"test_trace_expr(): self={self}")
        var = 4
        THE_MODULE.trace_expr(debug.get_level(), var, prefix="here: ")
        err = self.get_stderr()
        assert(my_re.search(r"here: var=4;?", err))

    @pytest.mark.xfail
    def test_multiline_trace_expr(self):
        """Make sure trace_expr expression resolved when split across lines"""
        var1 = 3
        var2 = 6
        THE_MODULE.trace_expr(debug.get_level(),
                              var1,
                              var2)
        err = self.get_stderr()
        assert(my_re.search(r"var1=3; var2=6;?", err))
        
    @pytest.mark.xfail
    def test_newline_trace_expr(self):
        """Test trace_expr with newline delim"""
        debug.trace(4, f"test_trace_expr(): self={self}")
        var1 = 3
        var2 = 6
        THE_MODULE.trace_expr(debug.get_level(), var1, var2, delim="\n")
        err = self.get_stderr()
        assert(my_re.search(r"var1=3\nvar2=6;?", err))

    @pytest.mark.xfail
    def test_trace_expr_max_len(self):
        """Test trace_expr with max_len"""
        debug.trace(4, f"test_trace_expr(): self={self}")
        var = "-" * 32
        THE_MODULE.trace_expr(debug.get_level(), var, max_len=8)
        err = self.get_stderr()
        ## TODO2: assert my_re.search(r"var='--------\.\.\.'", err))
        ##                                                 ^
        assert my_re.search(r"var='--------\.\.\.", err)

    @pytest.mark.xfail
    def test_trace_current_context(self):
        """Ensure trace_current_context works as expected"""
        debug.trace(4, f"test_trace_current_context(): self={self}")
        number: int = 9                 # pylint: disable=unused-variable
        self.patch_trace_level(4)
        THE_MODULE.trace_current_context(4)
        err = self.get_stderr()
        script_filename = gh.basename(__file__)
        # globals: {\n  {value): {\n
        # ... '__name__': 'mezcla.tests.test_debug',
        assert my_re.search(r"__name__[^,]+test_debug", err)
        # ... '__doc__': 'Tests for debug module', 
        assert my_re.search(r"__doc__[^,]+Tests for debug", err)
        # ... '__file__': '/home/joe/Mezcla/mezcla/tests/test_debug.py',
        assert my_re.search(fr"__file__[^,]+{script_filename}", err)
        # locals: {\n  {value): {\n
        # ...  'self': <mezcla.tests.test_debug.TestDebug testMethod=test_trace_current_context>,
        assert my_re.search(r"self[^,]+testMethod=test_trace_current_context", err)
        # ... 'number': 9}
        assert my_re.search(r"number[^,]+ 9", err)

    @pytest.mark.xfail
    def test_trace_exception(self):
        """Ensure trace_exception works as expected"""
        debug.trace(4, f"test_trace_exception(): self={self}")
        with pytest.raises(RuntimeError):
            raise RuntimeError("debug.trace failed")
        THE_MODULE.trace_exception(4, "debug.trace")
        err = self.get_stderr()
        assert "Exception during debug.trace" in err
        
    @pytest.mark.xfail
    def test_raise_exception(self):
        """Ensure raise_exception works as expected"""
        debug.trace(4, f"test_raise_exception(): self={self}")
        with pytest.raises(Exception):
            THE_MODULE.raise_exception()
        THE_MODULE.raise_exception(10)

    @pytest.mark.xfail
    def test_assertion(self):
        """Ensure assertion works as expected"""
        debug.trace(4, f"test_assertion(): self={self}")
        
        # Doesn't print in stderr
        THE_MODULE.assertion((2 + 2 + 1) == 5)
        err = self.get_stderr()
        assert 'failed' not in err
        
        # Prints assertion failed in stderr
        THE_MODULE.assertion((2 + 2) == 5)
        err = self.get_stderr()
        assert "failed" in err
        assert "(2 + 2) == 5" in err

        # Just traces the variable expression as is
        gpu = None
        THE_MODULE.assertion(gpu)
        err = self.get_stderr()
        ## ex: Assertion failed: gpu=None\n (at <ipython-input-14-e3c30214f890>:1)
        assert("Assertion failed: gpu (at" in err)

        
    @pytest.mark.xfail
    def test_multiline_assertion(self):
        """Make sure assertion expression split across lines resolved"""
        debug.trace(4, f"test_multiline_assertion(): self={self}")
        THE_MODULE.assertion(2 +
                             2 ==
                             5)
        err = self.get_stderr()
        self.do_assert(my_re.search(r"2.*\+.*2.*==.*5", err,
                                    flags=my_re.DOTALL|my_re.MULTILINE))

    def test_val(self):
        """Ensure val works as expected"""
        debug.trace(4, f"test_val(): self={self}")
        save_trace_level = THE_MODULE.get_level()
        test_value = 22
        THE_MODULE.set_level(5)
        level5_value = THE_MODULE.val(5, test_value)
        THE_MODULE.set_level(0)
        level0_value = THE_MODULE.val(1, test_value)
        THE_MODULE.set_level(save_trace_level)
        assert level5_value == test_value
        assert level0_value is None

    def test_code(self):
        """Ensure code works as expected"""
        debug.trace(4, f"test_code(): self={self}")
        ## TODO: debug.assertion(debug_level, debug.code(debug_level, lambda: (8 / 0 != 0.0)))
        save_trace_level = THE_MODULE.get_level()
        count = 0
        def increment():
            """Increase counter"""
            nonlocal count
            count += 1
        THE_MODULE.set_level(4)
        THE_MODULE.code(4, lambda: increment)
        THE_MODULE.set_level(save_trace_level)
        assert(count == 0)

    @pytest.mark.xfail
    def test_debug_print(self):
        """Ensure debug_print works as expected"""
        debug.trace(4, f"test_debug_print(): self={self}")
        self.monkeypatch.setattr("mezcla.debug.output_timestamps", True)

        THE_MODULE.debug_print('error foobar', -1)
        out,err  = self.get_stdout_stderr()
        assert "error foobar" in err
        assert not out

        # Test debug_file
        self.monkeypatch.setattr("mezcla.debug.debug_file", sys.stdout)
        THE_MODULE.debug_print('some text to test debug file', -1)
        out,err  = self.get_stdout_stderr()
        assert 'some text to test debug file' in out

    @pytest.mark.xfail
    def test_timestamp(self):
        """Ensure timestamp works as expected"""
        debug.trace(4, f"test_timestamp(): self={self}")
        debug_timestamp = THE_MODULE.timestamp()
        # example: 2025-02-02 01:23:27.451258
        assert my_re.search(r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d", debug_timestamp)

    def test_debugging(self):
        """Ensure debugging works as expected"""
        debug.trace(4, f"test_debugging(): self={self}")
        THE_MODULE.set_level(4)
        assert THE_MODULE.debugging(2)
        assert THE_MODULE.debugging(4)
        assert not THE_MODULE.debugging(6)

    def test_detailed_debugging(self):
        """Ensure detailed_debugging works as expected"""
        debug.trace(4, f"test_detailed_debugging(): self={self}")
        THE_MODULE.set_level(2)
        assert not THE_MODULE.detailed_debugging()
        THE_MODULE.set_level(4)
        assert THE_MODULE.detailed_debugging()
        THE_MODULE.set_level(6)
        assert THE_MODULE.detailed_debugging()

    def test_verbose_debugging(self):
        """Ensure verbose_debugging works as expected"""
        debug.trace(4, f"test_verbose_debugging(): self={self}")
        THE_MODULE.set_level(2)
        assert not THE_MODULE.verbose_debugging()
        THE_MODULE.set_level(5)
        assert THE_MODULE.verbose_debugging()
        THE_MODULE.set_level(7)
        assert THE_MODULE.verbose_debugging()

    def test_format_value(self):
        """Ensure format_value works as expected"""
        debug.trace(4, f"test_format_value(): self={self}")
        assert(THE_MODULE.format_value("\n    ", max_len=5) == "\\n   ...")
        assert(THE_MODULE.format_value("123456", max_len=7) == "123456")
        assert(THE_MODULE.format_value("123456", max_len=6) == "123456")
        assert(THE_MODULE.format_value("123456", max_len=5) == "12345...")
        assert(THE_MODULE.format_value("123456", max_len=4) == "1234...")
        assert(THE_MODULE.format_value("123456", max_len=3) == "123...")
        assert(THE_MODULE.format_value("123456", max_len=2) == "12...")
        assert(THE_MODULE.format_value("123456", max_len=1) == "1...")
        assert(THE_MODULE.format_value("123456", max_len=0) == "...")

    def test_format_value_strict(self):
        """Ensure format_value w/ strict works as expected"""
        debug.trace(4, f"test_format_value(): self={self}")
        def format_value_strict(value, max_len):
            """Invokes debug.format_value with strict option"""
            return THE_MODULE.format_value(value, max_len=max_len,
                                    strict=True)
        assert(format_value_strict("\n    ", max_len=5) == "\\n...")
        assert(format_value_strict("123456", max_len=7) == "123456")
        assert(format_value_strict("123456", max_len=6) == "123456")
        assert(format_value_strict("123456", max_len=5) == "12...")
        assert(format_value_strict("123456", max_len=4) == "1...")
        assert(format_value_strict("123456", max_len=3) == "...")
        assert(format_value_strict("123456", max_len=2) == "..")
        assert(format_value_strict("123456", max_len=1) == ".")
        assert(format_value_strict("123456", max_len=0) == "")

    def test_xor(self):
        """Ensure xor works as expected"""
        debug.trace(4, f"test_xor(): self={self}")
        THE_MODULE.set_level(7)
        # Test the XOR table
        assert not THE_MODULE.xor(0, 0.0)
        assert THE_MODULE.xor(0, 1.0)
        assert THE_MODULE.xor(1, 0.0)
        assert not THE_MODULE.xor(1, 1.0)
        # Test stderr
        err = self.get_stderr()
        assert "xor" in err

    def test_xor3(self):
        """Ensure xor3 works as expected"""
        debug.trace(4, f"test_xor3(): self={self}")
        THE_MODULE.set_level(7)
        # Test the XOR table
        assert not THE_MODULE.xor3(0, 0, 0)
        assert THE_MODULE.xor3(0, 0, 1)
        assert THE_MODULE.xor3(0, 1, 0)
        assert not THE_MODULE.xor3(0, 1, 1)
        assert THE_MODULE.xor3(1, 0, 0)
        assert not THE_MODULE.xor3(1, 0, 1)
        assert not THE_MODULE.xor3(1, 1, 0)
        assert not THE_MODULE.xor3(1, 1, 1)
        # Test stderr
        err = self.get_stderr()
        assert "xor3" in err

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_init_logging(self):
        """Ensure init_logging works as expected"""
        debug.trace(4, f"test_init_logging(): self={self}")
        
        # Test init_logging with env not GLOBAL_LOGGING
        self.monkeypatch.setenv("GLOBAL_LOGGING", 'False')
        THE_MODULE.init_logging()
        stderr_1 = self.get_stderr()
        assert "init_logging" in stderr_1
        assert "Setting root logger level " not in stderr_1
        
        # Test init_logging with env GLOBAL_LOGGING and detailed_debugging()
        self.clear_stdout_stderr()
        self.monkeypatch.setenv("GLOBAL_LOGGING", 'True')
        THE_MODULE.init_logging()
        new_level = THE_MODULE.logging.root.getEffectiveLevel()
        stderr_2 = self.get_stderr()
        assert "init_logging" in stderr_2
        assert "Setting root logger level " in stderr_2
        assert new_level == 10
        
        # Test init_logging with env GLOBAL_LOGGING and not detailed_debugging()
        self.clear_stdout_stderr()
        self.monkeypatch.setenv("GLOBAL_LOGGING", 'True')
        self.monkeypatch.setattr("mezcla.debug.trace_level", 3)
        THE_MODULE.init_logging()
        new_level = THE_MODULE.logging.root.getEffectiveLevel()
        assert new_level == 20

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_profile_function(self):
        """Ensure profile_function works as expected"""
        debug.trace(4, f"test_profile_function(): self={self}")
        self.monkeypatch.setattr("mezcla.debug.trace_level", 6)
        frame = inspect.currentframe()
        
        # test function call
        THE_MODULE.profile_function(frame, 'call', 'arg') 
        stderr = self.get_stderr()
        assert "in: mezcla.tests.test_debug:test_profile_function(arg);" in stderr
        
        # test function return
        self.clear_stderr()
        THE_MODULE.profile_function(frame, 'return', 'result') 
        stderr = self.get_stderr()
        assert "out: mezcla.tests.test_debug:test_profile_function => result" in stderr
        
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_reference_var(self):
        """Ensure reference_var works as expected"""
        debug.trace(4, f"test_reference_var(): self={self}")
        self.monkeypatch.setattr("mezcla.debug.trace_level", 10)
        THE_MODULE.reference_var("\'a\'")

        stderr = self.get_stderr()
        assert 'reference_var("\'a\'",)' in stderr

    def test_clip_value(self):
        """Ensure clip_value works as expected"""
        debug.trace(4, f"test_clip_value(): self={self}")
        assert THE_MODULE.clip_value('helloworld', 5) == 'hello...'
        assert THE_MODULE.clip_value('12345678910111213141516', 7) == '1234567...'

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_read_line(self):
        """Ensure read_line works as expected"""
        debug.trace(4, f"test_read_line(): self={self}")
        content = fix_indent(
            """
            line1
            line2
            line3
            """).lstrip("\n")
        
        temp_file = self.get_temp_file()
        system.write_file(temp_file, content)
        line_1 = THE_MODULE.read_line(temp_file, 1)
        line_2 = THE_MODULE.read_line(temp_file, 2)
        line_3 = THE_MODULE.read_line(temp_file, 3)
        assert my_re.search(fr"{line_1}.*{line_2}.*{line_3}", content)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_debug_init(self):
        """Ensure debug_init works as expected"""
        debug.trace(4, f"test_debug_init(): self={self}")
        self.patch_trace_level(6)
        temp_debug_filename = self.temp_file + ".debug.log"
        self.monkeypatch.setenv("DEBUG_FILE", temp_debug_filename)
        self.monkeypatch.setenv("ENABLE_LOGGING", "True")
        # NOTE: Setting MONITOR_FUNCTIONS to True breaks tests on windows
        today = str(datetime.now()).split(' ', maxsplit=1)[0]
        THE_MODULE.debug_init(force=True)
        # TODO3: why is _test*err.txt being output?
        err = self.get_stderr()
        err_file = system.form_path(gh.dirname(__file__),
                                    "_test_debug_init-err.txt")
        system.write_file(err_file, err)

        ## TODO2: use regex pattern matching to be less brittle
        assert f"debug_filename={temp_debug_filename}" in err
        assert f"debug_file=<_io.TextIOWrapper name=\'{temp_debug_filename}\'" in err
        assert f"[{THE_MODULE.__file__}] loaded at {today}" in err
        assert "Setting logger level to 10" in err
        assert f"DEBUG_FILE: {temp_debug_filename}" in err
        assert "ENABLE_LOGGING: True" in err

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_display_ending_time_etc(self):
        """Ensure display_ending_time_etc works as expected"""
        debug.trace(4, f"test_display_ending_time_etc(): self={self}")
        self.monkeypatch.setenv("SKIP_ATEXIT", "False")
        self.monkeypatch.delenv("DEBUG_FILE", raising=False)
        temp_debug_file = system.open_file(self.temp_file + ".debug.log",
                                           mode="w")
        assert(not temp_debug_file.closed)
        self.monkeypatch.setattr("mezcla.debug.debug_file", temp_debug_file)
        # note: display_ending_time_etc needs to be extracted from debug_init
        THE_MODULE.display_ending_time_etc()
        # TODO3: why is _test*err.txt being output?
        err = self.get_stderr()
        err_file = system.form_path(gh.dirname(__file__),
                                    "_test_display_ending_time_etc-err.txt")
        system.write_file(err_file, err)
        assert(temp_debug_file.closed)

    def test_visible_simple_trace(self):
        """Make sure level-1 trace outputs to stderr"""
        debug.trace(4, f"test_visible_simple_trace(): self={self}")
        self.setup_simple_trace()
        if not __debug__:
            self.expected_stderr_trace = ""
        pre_out, pre_err = self.get_stdout_stderr()
        save_trace_level = THE_MODULE.get_level()
        THE_MODULE.set_level(4)
        print(self.stdout_text)
        THE_MODULE.trace(3, self.stderr_text)
        THE_MODULE.set_level(save_trace_level)
        out, err = self.get_stdout_stderr()
        assert(self.expected_stdout_trace in out)
        assert(self.expected_stderr_trace in err)
        THE_MODULE.trace_expr(6, (pre_out, pre_err), (out,err))

    def test_hidden_simple_trace(self):
        """Make sure level-N+1 trace doesn't output to stderr"""
        debug.trace(4, f"test_hidden_simple_trace(): self={self}")
        self.setup_simple_trace()
        ## TEST
        ## capsys.stop_capturing()
        ## capsys.start_capturing()
        pre_out, pre_err = self.get_stdout_stderr()
        self.expected_stderr_trace = ""
        save_trace_level = THE_MODULE.get_level()
        THE_MODULE.set_level(0)
        print(self.stdout_text)
        THE_MODULE.trace(1, self.stderr_text)
        out, err = self.get_stdout_stderr()
        THE_MODULE.set_level(save_trace_level)
        assert self.expected_stdout_trace in out
        assert self.expected_stderr_trace in err
        THE_MODULE.trace_expr(6, (pre_out, pre_err), (out,err))



class TestDebug2(TestWrapper):
    """Another Class for test case definitions"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    
    def test_xor3_again(self):
        """Test xor3 again"""
        debug.trace(4, f"test_xor3_again(): self={self}")
        self.do_assert(debug.xor3(True, False, False))
        self.do_assert(not debug.xor3(True, True, True))
        self.do_assert(not debug.xor3(False, False, False))

    @pytest.mark.xfail
    def test_level(self):
        """"Make sure set_level honored (provided __debug__)"""
        debug.trace(4, f"test_level(): self={self}")
        old_level = debug.get_level()
        new_level = old_level + 1
        debug.set_level(new_level)
        expected_level = (new_level if __debug__ else old_level)
        level_set_ok = (debug.get_level() == expected_level)
        debug.set_level(old_level)
        self.do_assert(level_set_ok)

    @pytest.mark.xfail
    def test_trace_exceptions(self):
        """"Make sure debug.trace doesn't produce exceptions"""
        debug.trace(4, f"test_trace_exceptions(): self={self}")
        old_level = debug.get_level()
        debug.set_level(max(1, old_level))
        no_exception = True
        try:
            THE_MODULE.trace(1, 666)
        except:
            no_exception = False
            debug.trace_exception(5, "test_trace_exceptions")
        debug.set_level(old_level)
        self.do_assert(no_exception)

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Example for using the KenLM language modeling utility (based on sample):
#     http://kheafield.com/code/kenlm [./python/example.py]
# This assumes the language model has already been created via lmplz
#
#--------------------------------------------------------------------------------
# Note: Build steps for C-based utility (based on https://github.com/kpu/kenlm)
# - Requirements:
#   sudo apt install build-essential cmake libboost-system-dev libboost-thread-dev libboost-program-options-dev libboost-test-dev libeigen3-dev zlib1g-dev libbz2-dev liblzma-dev
# - Main code:
#   mkdir -p build
#   cd build
#   cmake ..
#   make -j 4
#   sudo make install
# - Python code:
#   pip install https://github.com/kpu/kenlm/archive/master.zip
# - For more info, see following:
#   https://github.com/kpu/kenlm
#   https://github.com/kpu/kenlm/blob/master/BUILDING
#--------------------------------------------------------------------------------
#
# TODO:
# - Add lint comment disabling extraneous lint warning (or modify python
#   interface to initialize it):
#     Module 'kenlm' has no 'LanguageModel' member (no-member)
#

"""Example for using the KenLM language modeling utility"""

# Standard modules
import os
import sys

# Installed modules
import kenlm

# Local modules
from mezcla import debug
from mezcla.system import getenv_boolean, getenv_text
from mezcla import glue_helpers as gh
from mezcla import tpo_common as tpo
from mezcla import system

# Constants
TL = debug.TL

# Environment options
# OLD: Initialize globals, including adhoc options via environment variables
DEFAULT_LM_FILE = os.path.join(os.path.dirname(__file__), '..', 'lm', 'test.arpa')
LM = getenv_text("LM", DEFAULT_LM_FILE)
SENT_DELIM = getenv_text("SENT_DELIM", "\n", "Delimiter for sentence splitting")
VERBOSE = getenv_boolean("VERBOSE", False, "Verbose output")

# Globals
## TEMP: Need to rework tests/test_kenkm_example.py
model = None
normalized_score = None

#-------------------------------------------------------------------------------

def summed_constituent_score(s):
    """Return sum of scores for ngrams for sentence S
    note: helper function used when checking that total full score equals direct score
    """
    global model
    return sum(prob for (prob, _len, _oov) in model.full_scores(s))
    

def main():
    """Entry point"""

    # Sanity check
    show_usage = (len(sys.argv) > 1) and ("--help" in sys.argv[1])
    if (not show_usage) and (not gh.non_empty_file(LM)):
        print("Unable to find usable language model file '%s'" % LM)
        show_usage = True
    if show_usage:
        print("Usage: %s [--help] [sentence | -]" % sys.argv[0])
        print("")
        print("Example:")
        print("  kenlm=~/programs/kenlm")
        print("  export PATH=$kenlm/bin:$PATH")
        print("  export LM=$kenlm/test.arpa")
        print("  %s" % __file__)
        print("")
        print("Notes:")
        print("- Use LM environment variable to specify alternative language model (see lmplz)")
        print("- Use SENT_DELIM to specify sentence delimiter (default is newline)")
        print("- To use standard input, specify - for sentence above.")
        print("- Use the following environment options to customize processing")
        print("\t" + tpo.formatted_environment_option_descriptions())
        ## OLD: sys.exit()
        ## TODO?: system.exit(status_code=0)
        system.exit()
    
    # Load model from ARPA-format file
    global model
    model = kenlm.LanguageModel(LM)
    print('{0}-gram model'.format(model.order))
    
    # Read input
    sentences = 'language modeling is fun .'
    if (len(sys.argv) > 1):
        sentences = " ".join(sys.argv[1:])
    if (sentences == "-"):
        sentences = sys.stdin.read()
    
    # Check each sentence (or phrase)
    # TODO: rework sto avoid reading sentences entirely into memory (e.g., via generic iterator)
    for sentence in sentences.split(SENT_DELIM):
        if not sentence.strip():
            continue
        print("sentence: %s" % sentence)
        print("model score: %s" % tpo.round_num(model.score(sentence)))
        global normalized_score
        normalized_score = (model.score(sentence) / len(sentence.split()))
        print("normalized score: %s" % tpo.round_num(normalized_score))
        gh.assertion(abs(summed_constituent_score(sentence) - model.score(sentence)) < 1e-3)
    
        # Print diagnostics in verbose mode
        if VERBOSE:
            words = ['<s>'] + sentence.split() + ['</s>']
    
            # Show scores and n-gram matches
            print("Constituent ngrams")
            ## OLD: print("Offset\tProb\tLength\tWords")
            ## OLD: for i, (prob, length) in enumerate(model.full_scores(sentence)):
            print("Offset\tProb\tLength\tUnseen\tWords")
            for i, (prob, length, unseen) in enumerate(model.full_scores(sentence)):
                start = i + 2 - length
                end = start + length
                ## OLD: print('{0}\t{1}\t{2}\t{3}'.format(i, tpo.round_num(prob), length, ' '.join(words[start : end])))
                unseen_spec = "*" if unseen else ""
                print('{0}\t{1}\t{2}\t{3}\t{4}'.format(i, tpo.round_num(prob), length, unseen_spec, ' '.join(words[start : end])))
    
            # Find out-of-vocabulary words
            print("out-of-vocabulary words: %s" % [w for w in words if (not w in model)])

#-------------------------------------------------------------------------------
            
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    debug.trace(5, f"module __doc__: {__doc__}")
    debug.assertion("TODO:" not in __doc__)
    main()



#! /usr/bin/env python3
#
# Tests for train_text_categorizer module
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_train_text_categorizer.py
#

"""Tests for train_text_categorizer module"""

# Standard modules
## NOTE: this is empty for now

# Installed modules
import pytest

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Rreference are used for the module to be tested:
#    THE_MODULE:	    global module object
system.setenv("USE_XGB", "1")
THE_MODULE = None
try:
    ## TEMP: fails if xgboost not available (workaround for stupid docker issue)
    import xgboost
    debug.trace_expr(5, xgboost.XGBClassifier)
    import mezcla.train_text_categorizer as THE_MODULE
except:
    system.print_exception_info("text_categorizer import")


@pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
class TestTrainTextCategorizer(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    resources = gh.resolve_path("resources")

    def test_usage(self):
        """Test usage statement"""
        # Current usage:
        #    Usage: ./train_text_categorizer.py training-file model-file [testing]
        #    ...
        #    - Currently, only tab-separated value (TSV) format is accepted:
        usage = self.run_script(options="--help")
        assert "model" in usage
        assert "TSV" in usage
    
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_data_file(self):
        """Makes sure TODO works as expected"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        data_file = gh.form_path(self.resources, "random-10pct-tweet-emotions.tsv")
        ## TODO: add use_stdin=True to following if no file argument
        output = self.run_script(options="", env_options="TEST_PERCENT=10 USE_XGB=1", data_file=data_file, post_options="-")
        assert my_re.search(r"Accuracy over.*: (\S+)", output.strip())
        accuracy = system.to_float(my_re.group(1))
        debug.trace_expr(5, accuracy)
        assert (accuracy > 0.20)
        return


if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)
